"""This module implements containers for storing pair counts
(:obj:`PatchedCount`) and the total number of objects (:obj:`PatchedTotal`) for
pair count normalisation. The data is stored per spatial patch and in bins of
redshift. The containers implement methods to compute total value (summing over
all patches) and samples needed for error estimations after evaluating the
correlation estimator (e.g. jackknife or bootstrap resampling).

Finally, :obj:`NormalisedCounts` implements normalised pair counts and holds
both a :obj:`PatchedCount` and :obj:`PatchedTotal` container. Its
:meth:`NormalisedCounts.sample` method computes the ratio of
counts-to-total-objects and samples thereof.
"""

from __future__ import annotations

import logging
from abc import abstractmethod
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from itertools import accumulate
from typing import TYPE_CHECKING, NoReturn, Union

try:  # pragma: no cover
    from typing import TypeAlias
except ImportError:  # pragma: no cover
    from typing_extensions import TypeAlias

import h5py
import numpy as np
import pandas as pd
from deprecated import deprecated

from yaw.config import ResamplingConfig
from yaw.core.abc import (
    BinnedQuantity,
    HDFSerializable,
    PatchedQuantity,
    concatenate_bin_edges,
)
from yaw.core.containers import Indexer, PatchIDs, SampledData
from yaw.core.logging import LogCustomWarning
from yaw.core.math import apply_slice_ndim, outer_triu_sum

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import DTypeLike, NDArray
    from pandas import IntervalIndex

__all__ = ["PatchedTotal", "PatchedCount", "NormalisedCounts"]


logger = logging.getLogger(__name__)

_compression = dict(fletcher32=True, compression="gzip", shuffle=True)
"""default compression settings for :obj:`h5py.Dataset`."""


TypeSlice: TypeAlias = Union[slice, int, None]
TypeIndex: TypeAlias = Union[int, slice, Sequence]


def check_mergable(patched_arrays: Sequence[PatchedArray], *, patches: bool) -> None:
    """Check if two instaces of PatchedArray can be merged along the patch
    or binning axis.

    Args:
        patched_arrays (:obj:`Sequence[PatchedArray]`):
            Instances to merge

    Keyword args:
        patches (:obj:`bool`):
            Whether to check for merging patches or binning.
    """
    reference = patched_arrays[0]
    for patched in patched_arrays[1:]:
        if reference.auto != patched.auto:
            raise ValueError("cannot merge mixed cross- and autocorrelations")
        if patches:
            reference.is_compatible(patched, require=True)
        else:
            if reference.auto != patched.auto:
                raise ValueError("cannot merge mixed cross- and autocorrelations")
            if reference.n_patches != patched.n_patches:
                raise ValueError("cannot merge, patch numbers do not match")


def binning_from_hdf(source: h5py.Group) -> IntervalIndex:
    """Construct a :obj:`pandas.IntervalIndex` from a group in an HDF5 file."""
    dset = source["binning"]
    left, right = dset[:].T
    closed = dset.attrs["closed"]
    return pd.IntervalIndex.from_arrays(left, right, closed=closed)


def binning_to_hdf(binning: IntervalIndex, dest: h5py.Group) -> None:
    """Serialise a :obj:`pandas.IntervalIndex` into a group of an HDF5 file.

    Stores the left and right edges for each interval, as well as on which side
    the intervals are closed as group attribute.
    """
    edges = np.column_stack([binning.left, binning.right])
    dset = dest.create_dataset("binning", data=edges, **_compression)
    dset.attrs["closed"] = binning.closed


def patch_idx_offset(patched: Iterable[PatchedArray]) -> NDArray[np.int_]:
    """Compute the offsets for patch indices of a set of :obj:`PatchedArray` if
    they were merged into one large :obj:`PatchedArray`."""
    idx_offset = np.fromiter(
        accumulate((p.n_patches for p in patched), initial=0),
        dtype=np.int_,
        count=len(patched),
    )
    return idx_offset


class PatchedArray(BinnedQuantity, PatchedQuantity, HDFSerializable):
    """Base class that implements the interface for classes that store results
    from pair count measurements.
    """

    auto = False
    """Whether the stored data are from an autocorrelation measurement."""

    def __repr__(self) -> str:
        string = super().__repr__()[:-1]
        shape = self.shape
        return f"{string}, {shape=})"

    @abstractmethod
    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            if self.n_bins != other.n_bins:
                return False
            elif self.n_patches != other.n_patches:
                return False
            elif not (self.get_binning() == other.get_binning()).all():
                return False
        return True

    def __neq__(self, other) -> bool:
        return not self == other

    @property
    def dtype(self) -> DTypeLike:
        """The numpy data type of the underlying data."""
        return np.float_

    @property
    def shape(self) -> tuple[int]:
        """The shape of underlying data if viewed as array."""
        return (self.n_patches, self.n_patches, self.n_bins)

    @property
    def ndim(self) -> int:
        """The number of dimensions of underlying data if viewed as array."""
        return len(self.shape)

    @property
    def size(self) -> int:
        """The number of items in the underlying data if viewed as array."""
        return np.prod(self.shape)

    @abstractmethod
    def as_array(self) -> NDArray:
        """Get the underlying data as contiguous array.

        The array 3-dimensional with shape (N, N, K), where N is the number of
        spatial patches, and K is the number of redshift bins."""
        raise NotImplementedError

    @abstractmethod
    def _sum(self, config: ResamplingConfig) -> NDArray:
        """Method that implements the sum over all patches."""
        raise NotImplementedError

    @abstractmethod
    def _jackknife(self, config: ResamplingConfig, signal: NDArray) -> NDArray:
        """Method that implements generating jackknife samples of the sum over
        all patches.

        For N patches, draw N realisations by leaving out one of the N patches.
        """
        raise NotImplementedError

    @abstractmethod
    def _bootstrap(self, config: ResamplingConfig) -> NDArray:
        """Method that implements generating bootstrap samples of the sum over
        all patches.

        For N patches, draw M realisations each containing N randomly chosen
        patches.
        """
        raise NotImplementedError

    @deprecated(reason="renamed to CorrFunc.sample_sum", version="2.3.1")
    def get_sum(self, *args, **kwargs):
        """
        .. deprecated:: 2.3.1
            Renamed to :meth:`sample_sum`.
        """
        return self.sample_sum(*args, **kwargs)

    def sample_sum(self, config: ResamplingConfig | None = None) -> SampledData:
        """Compute the sum of the data over all patches and samples thereof.

        Returns a data container with the sum in each redshift bin and samples
        generated from the patches using the resampling method specified in the
        configuration parameter.

        Args:
            config (:obj:`~yaw.config.ResamplingConfig`):
                Specifies the resampling method and its customisation
                parameters.

        Returns:
            :obj:`~yaw.core.SampledData`
        """
        if config is None:
            config = ResamplingConfig()  # pragma: no cover
        data = self._sum(config)
        if self.n_patches > 1:
            if config.method == "bootstrap":
                samples = self._bootstrap(config)
            else:
                samples = self._jackknife(config, signal=data)
        else:
            samples = np.atleast_2d(data)
        return SampledData(
            binning=self.get_binning(), data=data, samples=samples, method=config.method
        )


class PatchedTotal(PatchedArray):
    """Container class for the product of the total number of objects of two
    samples.

    The data in this container, the product of the total number of objects from
    two samples, is constructed by multiplting the total number of objects of
    both samples (split into spatial patches and redshift bins). This data is
    required to normalise pair counts when computing correlation functions, e.g.
    to account for different sizes of a data and a random sample.

    Internally, the nubmer of objects are stored per sample as :obj:`totals1`
    and :obj:`totals2` respectively. The (outer) product for all combinations
    of patches is only computed when calling the :meth:`as_array` or
    :meth:`sample_sum` methods.

    The container supports comparison of the data elements and the redshift
    binning with ``==`` and ``!=``. The indexing rules are the same as for
    :obj:`PatchedCount`.

    .. rubric:: Examples

    Select a subset of all redshift bins or all spatial patches:

    >>> from yaw.examples import patched_total
    >>> patched_total
    PatchedTotal(n_bins=30, z='0.070...1.420', shape=(64, 64, 30))

    Note how the indicated shape changes when a patch subset is selected:

    >>> patched_total.patches[:10]
    PatchedTotal(n_bins=30, z='0.070...1.420', shape=(10, 10, 30))

    Note how the indicated redshift range and shape change when a bin subset is
    selected:

    >>> patched_total.bins[:3]
    PatchedTotal(n_bins=3, z='0.070...0.205', shape=(64, 64, 3))

    An example of iteration over bins, which yields instances with a single
    redshift bin:

    >>> for zbin in patched_total.bins:
    ...     print(zbin)
    ...     break  # just show the first item
    PatchedTotal(n_bins=1, z='0.070...0.115', shape=(64, 64, 1))
    """

    totals1: NDArray
    """The total number of objects from the first data catalogue per patch and
    redshift bin.

    The array is of shape (N, K), where N is the number of spatial patches, and
    K is the number of redshift bins.
    """
    totals2: NDArray
    """The total number of objects from the second data catalogue per patch and
    redshift bin.

    The array is of shape (N, K), where N is the number of spatial patches, and
    K is the number of redshift bins.
    """

    def __init__(
        self, binning: IntervalIndex, totals1: NDArray, totals2: NDArray, *, auto: bool
    ) -> None:
        """Construct a new instance from the total number of objects in the
        first and second catalog.

        Args:
            binning (:obj:`pandas.IntervalIndex`):
                The redshift binning applied to the data.
            totals1 (:obj:`NDArray`):
                The total number of objects from the first data catalogue per
                patch and redshift bin. The array must be of shape (N, K), where
                N is the number of spatial patches, and K is the number of
                redshift bins.
            totals2 (:obj:`NDArray`):
                The total number of objects from the second data catalogue per
                patch and redshift bin. The array must be of shape (N, K), where
                N is the number of spatial patches, and K is the number of
                redshift bins.

        Keyword Args:
            auto (:obj:`bool`):
                Whether the data originates from an autocorrelation measurement.
        """
        self._binning = binning
        for i, totals in enumerate((totals1, totals2), 1):
            if totals.ndim != 2:
                raise ValueError(f"'totals{i}' must be two dimensional")
            if totals.shape[1] != self.n_bins:
                raise ValueError(
                    f"number of bins for 'totals{i}' does not match 'binning'"
                )
        if totals1.shape != totals2.shape:
            raise ValueError(
                f"number of patches and bins do not match: "
                f"{totals1.shape} != {totals2.shape}"
            )
        self.totals1 = totals1
        self.totals2 = totals2
        self.auto = auto

    def __eq__(self, other) -> bool:
        if not super().__eq__(other):
            return False  # checks type
        return (
            np.all(self.totals1 == other.totals1)
            and np.all(self.totals2 == other.totals2)
            and self.auto == other.auto
        )

    def as_array(self) -> NDArray:
        return np.einsum("i...,j...->ij...", self.totals1, self.totals2)

    @property
    def bins(self) -> Indexer[TypeIndex, PatchedTotal]:
        def builder(inst: PatchedTotal, item: TypeIndex) -> PatchedTotal:
            if isinstance(item, int):
                item = [item]
            return PatchedTotal(
                binning=inst._binning[item],
                totals1=inst.totals1[:, item],
                totals2=inst.totals2[:, item],
                auto=inst.auto,
            )

        return Indexer(self, builder)

    @property
    def patches(self) -> Indexer[TypeIndex, PatchedTotal]:
        def builder(inst: PatchedTotal, item: TypeIndex) -> PatchedTotal:
            if isinstance(item, int):
                item = [item]
            return PatchedTotal(
                binning=inst._binning,
                totals1=inst.totals1[item],
                totals2=inst.totals2[item],
                auto=inst.auto,
            )

        return Indexer(self, builder)

    def get_binning(self) -> IntervalIndex:
        return self._binning

    @property
    def n_patches(self) -> int:
        return self.totals1.shape[0]

    @classmethod
    def from_hdf(cls, source: h5py.Group) -> PatchedTotal:
        # reconstruct the binning
        binning = binning_from_hdf(source)
        # load the data
        totals1 = source["totals1"][:]
        totals2 = source["totals2"][:]
        auto = source["auto"][()]
        return cls(binning=binning, totals1=totals1, totals2=totals2, auto=auto)

    def to_hdf(self, dest: h5py.Group) -> None:
        # store the binning
        binning_to_hdf(self.get_binning(), dest)
        # store the data
        dest.create_dataset("totals1", data=self.totals1, **_compression)
        dest.create_dataset("totals2", data=self.totals2, **_compression)
        dest.create_dataset("auto", data=self.auto)

    def concatenate_patches(self, *data: PatchedTotal) -> PatchedTotal:
        all_totals: list[PatchedTotal] = [self, *data]
        check_mergable(all_totals, patches=True)
        return self.__class__(
            binning=self.get_binning().copy(),
            totals1=np.concatenate([t.totals1 for t in all_totals], axis=0),
            totals2=np.concatenate([t.totals2 for t in all_totals], axis=0),
            auto=self.auto,
        )

    def concatenate_bins(self, *data: PatchedTotal) -> PatchedTotal:
        all_totals: list[PatchedTotal] = [self, *data]
        check_mergable(all_totals, patches=False)
        binning = concatenate_bin_edges(*all_totals)
        return self.__class__(
            binning=binning,
            totals1=np.concatenate([t.totals1 for t in all_totals], axis=1),
            totals2=np.concatenate([t.totals2 for t in all_totals], axis=1),
            auto=self.auto,
        )

    # methods implementing the signal

    def _sum_cross(self) -> NDArray:
        """Implements the sum over all patches for crosscorrelation data.

        Effectively computes the sum of the outer product of :obj:`totals1` and
        :obj:`totals2` along the redshift bin axis.
        """
        return np.einsum("i...,j...->...", self.totals1, self.totals2)

    def _sum_auto(self) -> NDArray:
        """Implements the sum over all patches for autocorrelation data.

        The same as :meth:`_sum_cross`, but only computes the upper triangle of
        the outer product and weights the diagonal with 1/2 to account for the
        fact that the diagonal elements are true autocorrelations.
        """
        sum_upper = outer_triu_sum(self.totals1, self.totals2, k=1)
        sum_diag = np.einsum("i...,i...->...", self.totals1, self.totals2)
        return sum_upper + 0.5 * sum_diag

    def _sum(self, config: ResamplingConfig) -> NDArray:
        if self.auto:
            return self._sum_auto()
        else:
            return self._sum_cross()

    # methods implementing jackknife samples

    def _jackknife_cross(self, signal: NDArray) -> NDArray:
        """Implements jackknife samples of the sum over all patches for
        crosscorrelation data.

        The same as :meth:`_sum_cross`, but adds an additional dimension as
        first axis which lists the samples. For the i-th sample, subtract all
        data containing :obj:`totals1` or :obj:`totals2` from the i-th patch.
        """
        diag = np.einsum("i...,i...->i...", self.totals1, self.totals2)
        rows = np.einsum("i...,j...->i...", self.totals1, self.totals2)
        cols = np.einsum("i...,j...->j...", self.totals1, self.totals2)
        return signal - rows - cols + diag  # subtracted diag twice

    def _jackknife_auto(self, signal: NDArray) -> NDArray:
        """Implements jackknife samples of the sum over all patches for
        autocorrelation data.

        The same as :meth:`_jackknife_cross`, but only computes the upper
        triangle of the outer product for each sample and weights the diagonal
        with 1/2 to account for the fact that the diagonal elements are true
        autocorrelations.
        """
        diag = np.einsum("i...,i...->i...", self.totals1, self.totals2)
        # sum along axes of upper triangle (without diagonal) of outer product
        rows = outer_triu_sum(self.totals1, self.totals2, k=1, axis=1)
        cols = outer_triu_sum(self.totals1, self.totals2, k=1, axis=0)
        return signal - rows - cols - 0.5 * diag  # diag not in rows or cols

    def _jackknife(self, config: ResamplingConfig, signal: NDArray) -> NDArray:
        if self.auto:
            return self._jackknife_auto(signal)
        else:
            return self._jackknife_cross(signal)

    # methods implementing bootstrap samples

    def _bootstrap(self, config: ResamplingConfig, **kwargs) -> NoReturn:
        """Bootstrap resampling currently not implemented.

        Raises:
            :exc:`NotImplementedError`
        """
        raise NotImplementedError


class PatchedCount(PatchedArray):
    """Container class for pair counts between two samples.

    The data in this container are the pair counts between two samples of
    points. The counts are stored for each spatial patch and per redshift bin,
    forming a data array of shape shape (N, N, K), where N is the number of
    spatial patches, and K is the number of redshift bins.

    The container supports comparison of the data elements and the redshift
    binning with ``==`` and ``!=``. Additionally, :obj:`PatchedCount` can be
    added together if they have the same redshift binning and number of patches,
    e.g. to add pair counts measured on different scales. The count values can
    be rescaled/multiplied by a floating point number, e.g. to apply a weighting
    before summing different scales (see also
    :func:`yaw.correlation.add_corrfuncs`). Any sequence of :obj:`PatchedCount`
    can be summed together with the built-in python function ``sum()``.

    Finally, the container supports indexing of and iteration over redshift bins
    and spatial patches using the special accessor attributes :obj:`bins` (see
    also :obj:`~yaw.core.containers.SampledData`) and :obj:`patches`. Some
    examples are listed below.

    .. rubric:: Examples

    Create a redshift binning:

    >>> import pandas as pd
    >>> bins = pd.IntervalIndex.from_breaks([0.1, 0.2, 0.3])
    >>> bins
    IntervalIndex([(0.1, 0.2], (0.2, 0.3]], dtype='interval[float64, right]')

    Create two data containers with some dummy values:

    >>> count1 = PatchedCount.zeros(bins, n_patches=5, auto=False)
    >>> count1.counts += 1  # set all counts with dummy value 1
    >>> count2 = PatchedCount.zeros(bins, n_patches=5, auto=False)
    >>> count2.counts += 2  # set all counts with dummy value 2
    >>> count2
    PatchedCount(n_bins=2, z='0.100...0.300', shape=(5, 5, 2))

    Sum the pair counts and compare different methods:

    >>> summed = count1 + count2
    >>> summed
    PatchedCount(n_bins=2, z='0.100...0.300', shape=(5, 5, 2))
    >>> sum([count1, count2]) == summed
    True
    >>> (summed.counts == 3).all()
    True

    Rescale the pair counts:

    >>> count1 * 2.0
    PatchedCount(n_bins=2, z='0.100...0.300', shape=(5, 5, 2))
    >>> count1 * 2.0 == count2
    True

    Select a subset of all redshift bins or all spatial patches:

    >>> from yaw.examples import patched_count
    >>> patched_count
    PatchedCount(n_bins=30, z='0.070...1.420', shape=(64, 64, 30))

    Note how the indicated shape changes when a patch subset is selected:

    >>> patched_count.patches[:10]
    PatchedCount(n_bins=30, z='0.070...1.420', shape=(10, 10, 30))

    Note how the indicated redshift range and shape change when a bin subset is
    selected:

    >>> patched_count.bins[:3]
    PatchedCount(n_bins=3, z='0.070...0.205', shape=(64, 64, 3))

    An example of iteration over bins, which yields instances with a single
    redshift bin:

    >>> for zbin in patched_count.bins:
    ...     print(zbin)
    ...     break  # just show the first item
    PatchedCount(n_bins=1, z='0.070...0.115', shape=(64, 64, 1))
    """

    counts: NDArray
    """Internal data array containing the pair counts between spatial patches in
    bins of redshift.

    The array is 3-dimensional with shape (N, N, K), where N is the number of
    spatial patches, and K is the number of redshift bins. Same as
    :meth:`as_array`.
    """

    def __init__(
        self,
        binning: IntervalIndex,
        counts: NDArray,
        *,
        auto: bool,
    ) -> None:
        """Construct a new instance from an existing pair count array.

        Args:
            binning (:obj:`pandas.IntervalIndex`):
                The redshift binning applied to the data.
            counts (:obj:`NDArray`):
                Internal data array containing the pair counts between spatial
                patches in bins of redshift. The array must be 3-dimensional
                with shape (N, N, K), where N is the number of spatial patches,
                and K is the number of redshift bins. Same as :meth:`as_array`.

        Keyword Args:
            auto (:obj:`bool`):
                Whether the data originates from an autocorrelation measurement.
        """
        if counts.ndim != 3 or counts.shape[0] != counts.shape[1]:
            raise IndexError("counts must be of shape (n_patches, n_patches, n_bins)")
        if counts.shape[2] != len(binning):
            raise ValueError("length of 'binning' and 'counts' dimension do not match")
        self._binning = binning
        self.counts = counts
        self.auto = auto

    @classmethod
    def zeros(
        cls,
        binning: IntervalIndex,
        n_patches: int,
        *,
        auto: bool,
        dtype: DTypeLike = np.float_,
    ) -> PatchedCount:
        """Create a new instance where all elements of the counts array are
        initialised to zero.

        Args:
            binning (:obj:`pandas.IntervalIndex`):
                Redshift binning for the container, determines size of last data
                array dimension.
            n_patches (:obj:`int`):
                Number of spatial patches, determines the size of the first two
                data array dimensions.

        Keyword Args:
            auto (:obj:`bool`):
                Whether the data originates from an autocorrelation measurement.
            dtype (:obj:`DTypeLike`, optional):
                Data type to use for the internal data array.

        """
        counts = np.zeros((n_patches, n_patches, len(binning)), dtype=dtype)
        return cls(binning, counts, auto=auto)

    def __eq__(self, other) -> bool:
        if not super().__eq__(other):
            return False  # checks type
        return np.all(self.counts == other.counts) and (self.auto == other.auto)

    def __add__(self, other: PatchedCount) -> PatchedCount:
        self.is_compatible(other, require=True)
        if self.n_patches != other.n_patches:
            raise ValueError("number of patches does not agree")
        return self.__class__(
            self.get_binning(), self.counts + other.counts, auto=self.auto
        )

    def __radd__(self, other: PatchedCount | int | float) -> PatchedCount:
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __mul__(self, other: np.number) -> PatchedCount:
        return self.__class__(self.get_binning(), self.counts * other, auto=self.auto)

    def set_measurement(self, key: PatchIDs | tuple[int, int], item: NDArray):
        """Set the counts value in all redshift bins for a pair of patch
        indices.

        Args:
            key (:obj:`yaw.core.containers.PatchIDs`, tuple):
                Pair of patch indices for which the new values are set.
            item (:obj:`NDArray`):
                Values to set, must be an array with length matching the number
                of redshift bins.
        """
        # check the key
        if not isinstance(key, tuple):
            raise TypeError(f"slice must be of type {tuple}")
        elif len(key) != 2:
            raise IndexError(
                f"too many indices for array assignment: index must be "
                f"2-dimensional, but {len(key)} where indexed"
            )
        # check the item
        item = np.asarray(item)
        if item.shape != (self.n_bins,):
            raise ValueError(f"can only set items with length n_bins={self.n_bins}")
        # insert values
        self.counts[key] = item

    def as_array(self) -> NDArray:
        return self.counts

    def sum(self, axis: int | tuple[int] | None = None, **kwargs) -> NDArray:
        """Shorthand for :meth:`PatchedCount.counts.sum`

        Args:
            axis (:obj:`tuple`, :obj:`int`, optional):
                Axis over which the internal 3-dimensional data array is summed.
            **kwargs:
                Keyword arguments passed to :meth:`numpy.ndarry.sum`.
        """
        return self.counts.sum(axis=axis, **kwargs)

    @property
    def bins(self) -> Indexer[TypeIndex, PatchedCount]:
        def builder(inst: PatchedCount, item: TypeIndex) -> PatchedCount:
            if isinstance(item, int):
                item = [item]
            return PatchedCount(
                binning=inst._binning[item],
                counts=apply_slice_ndim(inst.counts, item, axis=2),
                auto=inst.auto,
            )

        return Indexer(self, builder)

    @property
    def patches(self) -> Indexer[TypeIndex, PatchedCount]:
        def builder(inst: PatchedCount, item: TypeIndex) -> PatchedCount:
            return PatchedCount(
                binning=inst._binning,
                counts=apply_slice_ndim(inst.counts, item, axis=(0, 1)),
                auto=inst.auto,
            )

        return Indexer(self, builder)

    def get_binning(self) -> IntervalIndex:
        return self._binning

    @property
    def n_patches(self) -> int:
        return self.counts.shape[0]

    @property
    def n_bins(self) -> int:
        return len(self.get_binning())

    def keys(self) -> NDArray:
        """Array of patch index pairs with non-zero pair counts.

        The index pairs are ordered by first, then second index. The returned
        array is of shape (N, 2), where N is the number patches that contain
        non-zero entries in any of the redshift bins.
        """
        has_data = np.any(self.counts, axis=2)
        indices = np.nonzero(has_data)
        return np.column_stack(indices)

    def values(self) -> NDArray:
        """Array of non-zero pair count values.

        The values are ordered in the same way as the indices returned by
        :meth:`keys`.
        """
        keys = self.keys()  # shape (n_nonzero, 2)
        i1, i2 = keys.T
        return self.counts[i1, i2, :]  # shape (n_nonzero, n_bins)

    @classmethod
    def from_hdf(cls, source: h5py.Group) -> PatchedCount:
        # reconstruct the binning
        binning = binning_from_hdf(source)
        # load the sparse data representation
        keys = [tuple(key) for key in source["keys"][:]]
        data = source["data"][:]
        n_patches = source["n_patches"][()]
        auto = source["auto"][()]
        # build dense counts matrix
        counts = np.zeros((n_patches, n_patches, len(binning)), dtype=data.dtype)
        for key, values in zip(keys, data):
            counts[key] = values
        return cls(binning=binning, counts=counts, auto=auto)

    def to_hdf(self, dest: h5py.Group) -> None:
        # store the binning
        binning_to_hdf(self.get_binning(), dest)
        # store the data
        dest.create_dataset("keys", data=self.keys(), **_compression)
        dest.create_dataset("data", data=self.values(), **_compression)
        dest.create_dataset("n_patches", data=self.n_patches)
        dest.create_dataset("auto", data=self.auto)

    def concatenate_patches(self, *data: PatchedCount) -> PatchedCount:
        all_counts: list[PatchedCount] = [self, *data]
        check_mergable(all_counts, patches=True)
        offsets = patch_idx_offset(all_counts)
        merged = self.__class__.zeros(
            binning=self.get_binning(),
            n_patches=sum(count.n_patches for count in all_counts),
            auto=self.auto,
        )
        # insert the blocks of counts into the merged counts array
        loc = 0
        for count, offset in zip(all_counts, offsets):
            merged.counts[loc : loc + offset, loc : loc + offset] = count.counts
            loc += offset
        return merged

    def concatenate_bins(self, *data: PatchedCount) -> PatchedCount:
        all_counts: list[PatchedCount] = [self, *data]
        check_mergable(all_counts, patches=False)
        binning = concatenate_bin_edges(*all_counts)
        merged = self.__class__.zeros(
            binning=binning, n_patches=self.n_patches, auto=self.auto, dtype=self.dtype
        )
        merged.counts = np.concatenate([count.counts for count in all_counts], axis=2)
        return merged

    # methods implementing the signal

    def _bin_sum_diag(self, data: NDArray) -> np.number:
        """Implements the sum over the autocorrelation patches in a single
        redshift bin.

        The autocorrelation patches are stored in the diagonal of the first two
        dimension of the pair count data array.

        .. Note::
            This method is valid for both cross- and autocorrelation data.
        """
        return np.diagonal(data).sum()

    def _bin_sum_cross(self, data: NDArray) -> np.number:
        """Implements the sum over all patches for crosscorrelation data in a
        single redshift bin.
        """
        return data.sum()

    def _bin_sum_auto(self, data: NDArray) -> np.number:
        """Implements the sum over all patches for autocorrelation data in a
        single redshift bin.

        The same as :meth:`_bin_sum_cross`, but only sums the upper triangle of
        the counts array.
        """
        return np.triu(data).sum()

    def _sum(self, config: ResamplingConfig) -> NDArray:
        out = np.zeros(self.n_bins)
        # compute the counts bin-wise
        for i in range(self.n_bins):
            data = self.counts[:, :, i]
            if config.crosspatch:
                if self.auto:
                    out[i] = self._bin_sum_auto(data)
                else:
                    out[i] = self._bin_sum_cross(data)
            else:
                out[i] = self._bin_sum_diag(data)
        return out

    # methods implementing jackknife samples

    def _bin_jackknife_diag(self, data: NDArray, signal: NDArray) -> NDArray:
        """Implements jackknife samples of the sum over the autocorrelation
        patches in a single redshift bin.

        The autocorrelation patches are stored in the diagonal of the first two
        dimension of the pair count data array. Samples are computed by leaving
        out the i-th patch in the i-th sample.

        .. Note::
            This method is valid for both cross- and autocorrelation data.
        """
        return signal - np.diagonal(data)  # broadcast to (n_patches,)

    def _bin_jackknife_cross(self, data: NDArray, signal: NDArray) -> NDArray:
        """Implements jackknife samples of the sum over all patches for
        crosscorrelation data in a single redshift bin.

        The same as :meth:`_bin_sum_cross`, but adds an additional dimension as
        first axis which lists the samples. For the i-th sample, subtract all
        data containing counts with objects of the i-th patch.
        """
        diag = np.diagonal(data)
        rows = data.sum(axis=1)
        cols = data.sum(axis=0)
        return signal - rows - cols + diag  # broadcast to (n_patches,)

    def _bin_jackknife_auto(self, data: NDArray, signal: NDArray) -> NDArray:
        """Implements jackknife samples of the sum over all patches for
        autocorrelation data in a single redshift bin.

        The same as :meth:`_bin_jackknife_cross`, but only sums elements on the
        main diagonal and upper triangle of the pair count array.
        """
        diag = np.diagonal(data)
        # sum along axes of upper triangle (without diagonal) of outer product
        tri_upper = np.triu(data, k=1)
        rows = tri_upper.sum(axis=1).flatten()
        cols = tri_upper.sum(axis=0).flatten()
        return signal - rows - cols - diag  # broadcast to (n_patches,)

    def _jackknife(self, config: ResamplingConfig, signal: NDArray) -> NDArray:
        out = np.empty((self.n_patches, self.n_bins))
        # compute the counts bin-wise
        for i, bin_signal in enumerate(signal):
            data = self.counts[:, :, i]
            if config.crosspatch:
                if self.auto:
                    out[:, i] = self._bin_jackknife_auto(data, bin_signal)
                else:
                    out[:, i] = self._bin_jackknife_cross(data, bin_signal)
            else:
                out[:, i] = self._bin_jackknife_diag(data, bin_signal)
        return out

    # methods implementing bootstrap samples

    def _bootstrap(self, config: ResamplingConfig, **kwargs) -> NoReturn:
        """Bootstrap resampling currently not implemented.

        Raises:
            :exc:`NotImplementedError`
        """
        raise NotImplementedError


@dataclass(frozen=True)
class NormalisedCounts(PatchedQuantity, BinnedQuantity, HDFSerializable):
    """Container to store counts and the total number of objects obtained from
    measuring pair counts for a correlation function.

    Both input containers must have the same binning and the same number of
    spatial patches. The container supports the same arithmetic as
    :obj:`PatchedCount` (see the listed examples), i.e. comparison, addition,
    multiplication by a scalar, as well as indexing. The resulting pair counts
    are always normalised by the number of objects stored in :obj:`total`.

    Args:
        count (:obj:`PatchedCount`):
            The pair count container.
        total (:obj:`PatchedTotal`):
            The container for the total number of objects from the samples.
    """

    count: PatchedCount
    """The pair count container."""
    total: PatchedTotal
    """The container for the total number of objects from the samples."""

    def __post_init__(self) -> None:
        if self.count.n_patches != self.total.n_patches:
            raise ValueError("number of patches of 'count' and total' do not match")
        if self.count.n_bins != self.total.n_bins:
            raise ValueError("number of bins of 'count' and total' do not match")

    def __repr__(self) -> str:
        string = super().__repr__()[:-1]
        n_patches = self.n_patches
        return f"{string}, {n_patches=})"

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return np.all(self.count == other.count) and np.all(
                self.total == other.total
            )
        else:
            return False

    def __neq__(self, other) -> bool:
        return not self == other

    def __add__(self, other: NormalisedCounts) -> NormalisedCounts:
        count = self.count + other.count
        if np.any(self.total.totals1 != other.total.totals1) or np.any(
            self.total.totals2 != other.total.totals2
        ):
            raise ValueError("total number of objects do not agree for operands")
        return self.__class__(count, self.total)

    def __radd__(self, other: NormalisedCounts | int | float) -> NormalisedCounts:
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def __mul__(self, other: np.number) -> NormalisedCounts:
        return self.__class__(self.count * other, self.total)

    @property
    def auto(self) -> bool:
        """Whether the stored data are from an autocorrelation measurement."""
        return self.count.auto

    @property
    def bins(self) -> Indexer[TypeIndex, NormalisedCounts]:
        def builder(inst: NormalisedCounts, item: TypeIndex) -> NormalisedCounts:
            if isinstance(item, int):
                item = [item]
            return NormalisedCounts(
                count=inst.count.bins[item], total=inst.total.bins[item]
            )

        return Indexer(self, builder)

    @property
    def patches(self) -> Indexer[TypeIndex, NormalisedCounts]:
        def builder(inst: NormalisedCounts, item: TypeIndex) -> NormalisedCounts:
            return NormalisedCounts(
                count=inst.count.patches[item], total=inst.total.patches[item]
            )

        return Indexer(self, builder)

    def get_binning(self) -> IntervalIndex:
        return self.total.get_binning()

    @property
    def n_patches(self) -> int:
        return self.total.n_patches

    def sample(self, config: ResamplingConfig) -> SampledData:
        """Sum the pair counts across all spatial patches and generate samples
        from resampling the patches.

        Calls the ``sample_sum()`` method of the :obj:`counts` and :obj:`total`
        containers. Returns a :obj:`~yaw.SampledData` instance that contains the
        normalised pair counts and spatially resampled values. Effectively
        computes :math:`\\frac{XX}{n_1 n_2}`, where :math:XX` are the number of
        pairs between the two samples with a total number of objects :math:`n_1`
        and :math:`n_2` respectively.

        Args:
            config (:obj:`~yaw.config.ResamplingConfig`):
                Specifies the resampling method and its customisation
                parameters.

        Returns:
            :obj:`~yaw.core.SampledData`
        """
        counts = self.count.sample_sum(config)
        totals = self.total.sample_sum(config)
        with LogCustomWarning(
            logger, "some patches contain no data after binning by redshift"
        ):
            samples = SampledData(
                binning=self.get_binning(),
                data=(counts.data / totals.data),
                samples=(counts.samples / totals.samples),
                method=config.method,
            )
        return samples

    @classmethod
    def from_hdf(cls, source: h5py.Group) -> NormalisedCounts:
        count = PatchedCount.from_hdf(source["count"])
        total = PatchedTotal.from_hdf(source["total"])
        return cls(count=count, total=total)

    def to_hdf(self, dest: h5py.Group) -> None:
        group = dest.create_group("count")
        self.count.to_hdf(group)
        group = dest.create_group("total")
        self.total.to_hdf(group)

    def concatenate_patches(self, *pcounts: NormalisedCounts) -> NormalisedCounts:
        counts = [pc.count for pc in pcounts]
        totals = [pc.total for pc in pcounts]
        return self.__class__(
            count=self.count.concatenate_patches(*counts),
            total=self.total.concatenate_patches(*totals),
        )

    def concatenate_bins(self, *pcounts: NormalisedCounts) -> NormalisedCounts:
        counts = [pc.count for pc in pcounts]
        totals = [pc.total for pc in pcounts]
        return self.__class__(
            count=self.count.concatenate_bins(*counts),
            total=self.total.concatenate_bins(*totals),
        )


def pack_results(
    count_dict: dict[str, PatchedCount], total: PatchedTotal
) -> NormalisedCounts | dict[str, NormalisedCounts]:
    """Pack pair counts and the total number of objects.

    If measured for multiple scales, the counts should be a dictionary of with
    the scale name as key. In this case, the function returns a dictionary of
    :obj:`NormalisedCounts` with the same keys.
    """
    # drop the dictionary if there is only one scale
    if len(count_dict) == 1:
        count = tuple(count_dict.values())[0]
        result = NormalisedCounts(count=count, total=total)
    else:
        result = {}
        for scale_key, count in count_dict.items():
            result[scale_key] = NormalisedCounts(count=count, total=total)
    return result



"""This module implements helper classes to handle parallel computation with
:mod:`multiprocessing`.
"""

from __future__ import annotations

import functools
import logging
import multiprocessing
from collections.abc import Callable, Collection, Iterable, Iterator
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, TypeVar

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import DTypeLike, NDArray

__all__ = ["SharedArray", "ParallelHelper"]


logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class SharedArray:
    """Container for shared array that can be easily shared with other
    processes.

    Stores multidimenional data boolean, integer or floating point numbers in a
    flat array in shared memory. The recommended constructors are :meth:`empty`
    and :meth:`from_numpy`.

    Args:
        array (:obj:`multiprocessing.RawArray`):
            A raw array of shared memory used to store array data.
        shape (:obj:`tuple[int]`):
            Shape of the array.
        dtype (:obj:`DTypeLike`):
            Numpy-style data type of the array.
    """

    array: multiprocessing.RawArray
    """Underlying shared memory array."""
    shape: tuple[int]
    """Shape of the array."""
    dtype: DTypeLike
    """Numpy-style data type of the array."""

    @staticmethod
    def _get_mp_type(dtype: DTypeLike | str) -> str:
        type_map = {
            "|b1": "b",  # no bool correspondance, but same number of bytes
            "i1": "h",
            "u1": "H",
            "i2": "i",
            "u2": "I",
            "i4": "l",
            "u4": "L",
            "i8": "q",
            "u8": "Q",
            "f4": "f",
            "f8": "d",
        }
        if isinstance(dtype, str):
            dtype = np.dtype(dtype)
        try:
            return type_map[dtype.str.strip("<>")]
        except KeyError as e:
            raise NotImplementedError(
                f"cannot convert numpy dtype '{dtype.str}' to RawArray"
            ) from e

    def __len__(self) -> int:
        return self.shape[0]

    def __del__(self) -> None:
        object.__delattr__(self, "array")

    @classmethod
    def empty(cls, shape: tuple[int], dtype: DTypeLike | str) -> SharedArray:
        """Create an new class instance with uninitialised memory.

        Args:
            shape (:obj:`tuple[int]`):
                Shape of the array.
            dtype (:obj:`DTypeLike`):
                Numpy-style data type of the array.

        Returns:
            :obj:`SharedArray`
        """
        mtype = cls._get_mp_type(dtype)
        size = int(np.prod(shape))
        new = cls(
            multiprocessing.RawArray(mtype, size), shape=shape, dtype=np.dtype(dtype)
        )
        return new

    @classmethod
    def from_numpy(cls, array: NDArray) -> SharedArray:
        """Create an new class instance from a numpy array.

        Args:
            array (:obj:`NDArray`):
                Input array, used to initialise the shared memory.

        Returns:
            :obj:`SharedArray`
        """
        new = cls.empty(array.shape, array.dtype)
        try:
            buffer = new.to_numpy()
            np.copyto(buffer, array)
            return new
        except Exception:
            del new
            raise

    def to_numpy(self, copy: bool = False) -> NDArray:
        """Reconstruct the original numpy array from shared memory.

        Args:
            copy (:obj:`bool`):
                Whether to copy the array values instead of wrapping the shared
                memory (zero-copy).

        Returns:
            :obj:`NDArray`
        """
        arr = np.frombuffer(self.array, dtype=self.dtype).reshape(self.shape)
        if copy:
            return np.copy(arr)
        else:
            return arr


POOL_SHARE = dict()
"""Global dictionary used to share :obj:`SharedArray` instances with other
processes."""


def _threadinit(
    shares_dict: dict[str, SharedArray],
    initializer: Callable | None = None,
    initargs: Iterable | None = None,
) -> None:
    """Initialiser function used to send shared arrays to processes."""
    POOL_SHARE.update(shares_dict)
    # call the usual initializer function provided by the user
    if initializer is not None:
        initializer(*initargs)


def _threadwrapper(arg_tuple, function):
    """Simple wrapper function that unpacks a tuple of arguments and applies to
    a function."""
    return function(*arg_tuple)


_T = TypeVar("_T")


class ParallelHelper(Generic[_T]):
    """Helper class to run a function in parallel with
    :obj:`multiprocessing.Pool`.

    The positional function arguments are registered in order of appearance.
    Arguments can be either iterable (all of the same length), which will be
    iterated for each parallel function call, or constant, which are identical
    at each function call.

    Additionally, :obj:`SharedArray` instances can be provided which will be
    distributed automatically.

    .. Warning:
        Calling the target function with keyword arguments is not supported.

    .. Tip::
        After completion, call the :meth:`free` method to free the memory of all
        shared array, or use this class as a context wrapper, which will
        automate this process on exit.
    """

    def __init__(
        self, function: Callable[..., _T], n_items: int, num_threads: int | None = None
    ) -> None:
        """Create a new instane.

        Args:
            function (Callable):
                Function which accepts an arbitray number of positional
                arguments.
            n_items (:obj:`int`):
                The total number of items/jobs to apply to the function.
            num_threads (:obj:`int`):
                The total number of threads to use for parallel processing.
                Capped at the maximum number of logical CPU cores.
        """
        self.function = function
        self._n_items = n_items
        if num_threads is None:
            num_threads = multiprocessing.cpu_count()
        self._num_threads = num_threads if n_items > num_threads else n_items
        self.shares = {}
        self.args = []

    def __enter__(self) -> ParallelHelper:
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> None:
        self.free()

    def free(self) -> None:
        """Free the memory of all shared arrays registered with
        :meth:`add_shared_array`."""
        for key in list(self.shares.keys()):
            sarray = self.shares.pop(key)
            del sarray

    def n_jobs(self) -> int:
        """The length of the iterable function arguments, i.e. the total number
        of jobs."""
        return self._n_items

    def n_args(self) -> int:
        """The total number of arguments passed at each function call."""
        return len(self.args)

    def n_threads(self):
        """The total number of threads to use for processing."""
        return self._num_threads

    def add_shared_array(self, key: str, array: NDArray | SharedArray) -> None:
        """Add a new shared array as function argument.

        .. Note::
            The array will be treated as constant, i.e. the same array is
            applied at each function call.

        Args:
            key (:obj:`str`):
                Identifier name at which the array is registered in the new
                processes namespace.
            array (:obj:`NDArray`, :obj:`SharedArray`):
                Array, if numpy array, converted automatically to shared array.
        """
        if isinstance(array, SharedArray):
            sarray = array
        else:
            sarray = SharedArray.from_numpy(array)
        self.shares[key] = sarray

    def add_constant(self, value: Any) -> None:
        """Add the next function argument which will be repeated at each
        function call."""
        self.args.append([value] * self._n_items)

    def add_iterable(self, iterable: Collection) -> None:
        """Add the next function argument which will be iterated at each
        function call.

        Must be an iterable with :meth:`n_jobs` items.
        """
        if len(iterable) != self._n_items:
            raise ValueError(f"length of iterable argument must be {self._n_items}")
        self.args.append(iterable)

    def _init_pool(
        self, initializer: Callable | None = None, initargs: Iterable | None = None
    ) -> multiprocessing.Pool:
        """Initialises the worker processes for the :obj:`multiprocessing.Pool`."""
        all_initargs = [self.shares]
        if initializer is not None:
            all_initargs.append(initializer)
            if isinstance(initargs, Iterable):
                all_initargs.append(initargs)
            elif initargs is None:
                all_initargs.append(None)
            else:
                raise TypeError("'initargs' must be an iterable")
        logger.debug(f"running {self.n_jobs()} jobs on {self.n_threads()} threads")
        return multiprocessing.Pool(
            initializer=_threadinit, initargs=all_initargs, processes=self._num_threads
        )

    def result(
        self, initializer: Callable | None = None, initargs: Iterable | None = None
    ) -> list[_T]:
        """Apply the accumulated arguments to a function in a pool of processes.

        The call is blocking until all results are received.

        Args:
            initializer (Callable, optional):
                Custom initialiser function to call when creating the pool.
            initargs (Iterable, optional):
                Arguments applied to the initialiser function.

        Returns:
            :obj:`list`:
                List containing the return values from all function calls.
        """
        with self._init_pool(initializer, initargs) as pool:
            results = pool.starmap(self.function, zip(*self.args))
        return results

    def iter_result(
        self,
        initializer: Callable | None = None,
        initargs: Iterable | None = None,
        ordered: bool = True,
    ) -> Iterator[_T]:
        """Apply the accumulated arguments to a function in a pool of processes.

        Returns an iterator over the available thread results.

        Args:
            initializer (Callable, optional):
                Custom initialiser function to call when creating the pool.
            initargs (Iterable, optional):
                Arguments applied to the initialiser function.

        Yields:
            :obj:`list`:
                Iterator over the return values.
        """
        function = functools.partial(_threadwrapper, function=self.function)
        with self._init_pool(initializer, initargs) as pool:
            imap_args = (function, zip(*self.args))
            if ordered:
                for result in pool.imap(*imap_args):
                    yield result
            else:
                for result in pool.imap_unordered(*imap_args):
                    yield result



"""This module defines an interface for custom cosmological models, as well as
routines that depend on cosmological distance calculations.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

try:  # pragma: no cover
    from typing import TypeAlias
except ImportError:  # pragma: no cover
    from typing_extensions import TypeAlias

import numpy as np
from astropy.cosmology import FLRW, Planck15

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import ArrayLike, NDArray

__all__ = [
    "get_default_cosmology",
    "CustomCosmology",
    "r_kpc_to_angle",
    "Scale",
    "BinFactory",
]


def get_default_cosmology() -> FLRW:
    """Get the default cosmology (Planck Collaboration et al. 2015)."""
    return Planck15


class CustomCosmology(ABC):
    """Metaclass that defines the API to implement a custom cosmological model.

    The two required methods should behave like the corresponding methods in
    :obj:`astropy.cosmology.FLRW`.
    """

    @abstractmethod
    def comoving_distance(self, z: ArrayLike) -> ArrayLike:
        """Comoving line-of-sight distance in Mpc at a given redshift.

        The comoving distance along the line-of-sight between two objects
        remains constant with time for objects in the Hubble flow.

        Args:
            z (Quantity-like ``redshift``, :obj:`NDArray`, or scalar number):
                Input redshift.

        Returns:
            :obj:`astropy.units.Quantity`:
                Comoving distance in Mpc to each input redshift.
        """
        NotImplemented

    @abstractmethod
    def comoving_transverse_distance(self, z: ArrayLike) -> ArrayLike:
        """Comoving transverse distance in Mpc at a given redshift.

        This value is the transverse comoving distance at redshift :math:`z`
        corresponding to an angular separation of 1 radian. This is the same as
        the comoving distance if :math:`\\Omega_k` is zero (as in the current
        concordance Lambda-CDM model).

        Args:
            z (Quantity-like ``redshift``, :obj:`NDArray`, or scalar number):
                Input redshift.

        Returns:
            :obj:`astropy.units.Quantity`:
                Comoving transverse distance in Mpc at each input redshift.
        """
        NotImplemented


TypeCosmology: TypeAlias = Union[FLRW, CustomCosmology]


def r_kpc_to_angle(
    r_kpc: NDArray[np.float_] | Sequence[float], z: float, cosmology: TypeCosmology
) -> NDArray[np.float_]:
    """Convert from a physical separation in kpc to angles in radian at a given
    redshift.

    Args:
        r_kpc (:obj:`NDArray`, number):
            Physical separation in kpc.
        z (:obj:`float`):
            Redshift at which conversion happens.
        cosmology (:obj:`astropy.cosmology.FLRW`, :obj:`CustomCosmology`):
            Cosmological model used for calculations.

    Returns:
        :obj:`NDArray`:
            Angular separation at given redshift.
    """
    f_K = cosmology.comoving_transverse_distance(z)  # for 1 radian in Mpc
    return np.asarray(r_kpc) / 1000.0 * (1.0 + z) / f_K.value


@dataclass(frozen=True)
class Scale:
    """Class that represents a range of physical scales in kpc.

    The range is defined by a lower and an upper scale limit. An instance can be
    converted to a string representation that is used as dictionary key in some
    places of ``yaw``.

    .. rubric:: Examples

    Get the center points:

    >>> scale = Scale(100, 1000)
    >>> scale.mid, scale.mid_log
    (550.0, 316.22776601683796)

    String representation:

    >>> str(scale)
    'kpc100t1000'
    """

    rmin: float
    """Lower scale limit in kpc."""
    rmax: float
    """Upper scale limit in kpc."""

    def __post_init__(self) -> None:
        if self.rmin >= self.rmax:
            raise ValueError("'rmin' must be less than 'rmax'")

    @property
    def mid(self) -> float:
        """The mid point of the scale range."""
        return (self.rmin + self.rmax) / 2.0

    @property
    def mid_log(self) -> float:
        """The logarithmic (base 10) mid point of the scale range."""
        lmin = np.log10(self.rmin)
        lmax = np.log10(self.rmax)
        lmid = (lmin + lmax) / 2.0
        return 10**lmid

    def __str__(self) -> str:
        return f"kpc{self.rmin:.0f}t{self.rmax:.0f}"

    def to_radian(self, z: float, cosmology: TypeCosmology) -> NDArray[np.float_]:
        """Get the separation in radian at a given redshift.

        Args:
            z (:obj:`float`):
                Redshift at which conversion happens.
            cosmology (:obj:`astropy.cosmology.FLRW`, :obj:`CustomCosmology`):
                Cosmological model used for calculations.

        Returns:
            :obj:`NDArray`:
                Angular separation at given redshift.
        """
        return r_kpc_to_angle([self.rmin, self.rmax], z, cosmology)


class BinFactory:
    """Class used to generate redshift bins."""

    def __init__(
        self,
        zmin: float,
        zmax: float,
        nbins: int,
        cosmology: TypeCosmology | None = None,
    ):
        """Create a new bin generator.

        Args:
            zmin (:obj:`float`):
                Minimum redshift, lowest redshift bin edges.
            zmax (:obj:`float`):
                Maximum redshift, lowest redshift bin edges.
            nbins (:obj:`int`):
                Number of bins to generate.
            cosmology (:obj:`astropy.cosmology.FLRW`, :obj:`CustomCosmology`):
                Cosmological model used for calculations.
        """
        if zmin >= zmax:
            raise ValueError("'zmin' >= 'zmax'")
        if cosmology is None:
            cosmology = get_default_cosmology()
        self.cosmology = cosmology
        self.zmin = zmin
        self.zmax = zmax
        self.nbins = nbins

    def linear(self) -> NDArray[np.float_]:
        """Generate a binning with equal width in redshift."""
        return np.linspace(self.zmin, self.zmax, self.nbins + 1)

    def comoving(self) -> NDArray[np.float_]:
        """Generate a binning with equal width in radial comoving distance."""
        cbinning = np.linspace(
            self.cosmology.comoving_distance(self.zmin).value,
            self.cosmology.comoving_distance(self.zmax).value,
            self.nbins + 1,
        )
        # construct a spline mapping from comoving distance to redshift
        zarray = np.linspace(0, 10.0, 5000)
        carray = self.cosmology.comoving_distance(zarray).value
        return np.interp(cbinning, xp=carray, fp=zarray)  # redshift @ cbinning

    def logspace(self) -> NDArray[np.float_]:
        """Generate a binning with equal width in logarithmic redshift
        :math:`\\log(1+z)`."""
        logbinning = np.linspace(
            np.log(1.0 + self.zmin), np.log(1.0 + self.zmax), self.nbins + 1
        )
        return np.exp(logbinning) - 1.0

    @staticmethod
    def check(zbins: NDArray[np.float_]) -> None:
        """Check if a list of bin edges in monotonicaly increasing.

        Raises a :exc:`ValueError` if the condition is not met.

        Args:
            zbins (:obj:`NDArray`):
                Redshift bin edges to check.
        """
        if np.any(np.diff(zbins) <= 0):
            raise ValueError("redshift bins must be monotonicaly increasing")

    def get(self, method: str) -> NDArray[np.float_]:
        """Call one of the generation methods based on its name.

        Args:
            method (:obj:`str`):
                Name of the generation method.
        """
        try:
            return getattr(self, method)()
        except AttributeError as e:
            raise ValueError(f"invalid binning method '{method}'") from e



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from scipy.spatial import cKDTree

from yaw.core.coordinates import Coordinate, DistSky

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray

__all__ = ["SphericalKDTree"]


class InvalidScalesError(Exception):
    pass


class SphericalKDTree:
    """Wrapper around :obj:`scipy.spatial.cKDTree` that represents angular
    coordinates as points on the unitsphere.

    The only implemented operation is counting pairs in a fixed angular annulus.
    Angular distances are converted to the corresponding Euclidean distance on
    the unitsphere. Individual weights for points are supported.
    """

    _total = None

    def __init__(
        self,
        position: Coordinate,
        weights: NDArray[np.float_] | None = None,
        leafsize: int = 16,
    ) -> None:
        """Build a new tree from a set of coordinates.

        Args:
            position (:obj:`yaw.coordinates.Coordinate`):
                A vector of coordinates in either angular or 3D coordiantes, is
                converted to 3D coordinates if needed.
            weights (:obj:`NDArray`, optional):
                Individual weights for the points.
            leafsize (:obj:`int`, optional):
                Size at which branches of the KDTree are considered leaf nodes
                with no further childs.
        """
        position = np.atleast_2d(position.to_3d().values)
        self.tree = cKDTree(position, leafsize)
        if weights is None:
            self.weights = np.ones(len(position))
        else:
            assert len(weights) == len(position)
            self.weights = np.asarray(weights)

    def __len__(self) -> int:
        return len(self.weights)

    @property
    def total(self) -> float:
        """Sum of weights or total number of objects if not provided."""
        if self._total is None:
            self._total = self.weights.sum()
        return self._total

    def count(
        self,
        other: SphericalKDTree,
        scales: NDArray[np.float_],
        dist_weight_scale: float | None = None,
        weight_res: int = 50,
    ) -> NDArray:
        """Count pairs on a set of angular scales.

        Pairs are counted with in a range of minimum and maximum angle in
        radian. If multiple scales are provided, the set of scales is converted
        into a list of radial bins. After counting, the binned counts are summed
        to obtain the counts for the (potentially overlapping) input scales.

        The method also supports weighting the pairs radially by a simple
        power-law :math:`r^\\alpha`, where :math:`r` is the pair separation. To
        speed up computation, the weight is computed individually, but for all
        pairs within one angular bin in the logarithmic center of the bin. If
        radial weights are provided, the resultion of the angular binning is
        increased beyond the binning obtained by combining the scale limits (see
        above).

        Args:
            other (:obj:`SphericalKDTree`):
                Second tree used to count pairs.
            scales (:obj:`NDArray`):
                Array with angular scales in radian with shape (2, N). The
                scales are provided as at least one tuple of minimum and maximum
                angular scale.
            dist_weight_scale (:obj:`float`, optional):
                The power-law index for the radial weighting.
            weight_res (:obj:`NDArray`):
                The number of logarithmic angular bins used to compute the
                angular weights. Ignored if no power-law index is set.

        Returns:
            :obj:`NDArray`:
                The pair counts for each input scale, with optional inidividual
                point weights and radial weights applied.

        .. Warning::

            For autocorrelation measurements, ``other`` must be the same
            tree as the calling instance itself. This will results in pairs
            being counted twice, as they normally would be in the cross-tree
            counting case.
        """
        # unpack query scales
        scales = np.atleast_2d(scales)
        if scales.shape[1] != 2:
            raise InvalidScalesError("'scales' must be composed of tuples of length 2")
        if np.any(scales <= 0.0):
            raise InvalidScalesError("scales must be positive (r > 0)")
        if np.any(scales > np.pi):
            raise InvalidScalesError("scales exceed 180 deg")
        log_scales = np.log10(scales).flatten()
        # construct bins
        rlog_edges = np.linspace(log_scales.min(), log_scales.max(), weight_res)
        rlog_edges = np.array(sorted(set(rlog_edges) | set(log_scales)))
        r_edges = 10**rlog_edges
        # count pairs
        try:
            counts = self.tree.count_neighbors(
                other.tree,
                DistSky(r_edges).to_3d().values,
                weights=(self.weights, other.weights),
                cumulative=False,
            )
        except IndexError:
            counts = np.zeros_like(r_edges)
        counts = counts[1:]  # discard counts with 0 < R <= r_min
        # apply the distance weights
        if dist_weight_scale is not None:
            rlog_centers = (rlog_edges[:-1] + rlog_edges[1:]) / 2.0
            counts *= (10**rlog_centers) ** dist_weight_scale
        # compute counts for original bins
        result = np.empty(len(scales))
        for i, scale in enumerate(scales):
            i_lo = np.argmin(np.abs(r_edges - scale[0]))
            i_hi = np.argmin(np.abs(r_edges - scale[1]))
            select = np.arange(i_lo, i_hi)
            result[i] = counts[select].sum()
        return result



"""This module defines a few containers used throughout other modules.

Most importantly, they implement two containers for data with attached samples
(e.g. jackknife or bootstrap). Scalar values are implemented in
:obj:`SampledValue`, data that has been computed from redshift bins in
:obj:`SampledData`, which also serves as base class for most other containers
with redshift binning.
"""

from __future__ import annotations

import warnings
from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING, Callable, Generic, NamedTuple, TypeVar

import numpy as np
import pandas as pd

from yaw.config import OPTIONS
from yaw.core.abc import BinnedQuantity, concatenate_bin_edges
from yaw.core.math import cov_from_samples

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray
    from pandas import DataFrame, IntervalIndex, Series

__all__ = ["Indexer", "PatchIDs", "PatchCorrelationData", "SampledValue", "SampledData"]


_TK = TypeVar("_TK")
_TV = TypeVar("_TV")


class Indexer(Generic[_TK, _TV], Iterator):
    """Helper class to implemented a class attribute that can be used as
    indexer and iterator for the classes stored data (e.g. indexing patches or
    redshift bins).
    """

    def __init__(self, inst: _TV, builder: Callable[[_TV, _TK], _TV]) -> None:
        """Construct a new indexer.

        Args:
            inst:
                Class instance on which the indexing operations are applied.
            builder:
                Callable signature ``builder(inst, item) -> inst`` that
                constructs a new class instance with the indexing specified from
                ``item`` applied.


        The resulting indexer supports indexing and slicing (depending on the
        subclass implementation), as well as iteration, where instances holding
        individual items are yielded.
        """
        self._inst = inst
        self._builder = builder
        self._iter_loc = 0

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._inst.__class__.__name__})"

    def __getitem__(self, item: _TK) -> _TV:
        return self._builder(self._inst, item)

    def __next__(self) -> _TV:
        """Returns the next value and increments the iterator location index."""
        try:
            val = self[self._iter_loc]
        except IndexError:
            raise StopIteration
        else:
            self._iter_loc += 1
            return val

    def __iter__(self) -> Iterator[_TV]:
        """Returns a new instance of this class to have an independent iterator
        location index"""
        return self.__class__(inst=self._inst, builder=self._builder)


class PatchIDs(NamedTuple):
    """Named tuple that can hold a pair of patch indices."""

    id1: int
    """First patch index."""
    id2: int
    """Second patch index."""


@dataclass(frozen=True)
class PatchCorrelationData:
    """Container to hold the result of a pair counting operation between two
    spatial patches.

    Args:
        patches (:obj:`PatchIDs`):
            The indices of used the patches.
        totals1 (:obj:`NDArray`):
            Total number of objects after binning by redshift in first patch.
        totals1 (:obj:`NDArray`):
            Total number of objects after binning by redshift in second patch.
        counts (:obj:`dict`):
            Dictionary listing the number of counted pairs after binning by
            redshift. Each item represents results from a different scale.
    """

    patches: PatchIDs
    totals1: NDArray
    totals2: NDArray
    counts: dict[str, NDArray]


_Tscalar = TypeVar("_Tscalar", bound=np.number)


@dataclass(frozen=True)
class SampledValue(Generic[_Tscalar]):
    """Container to hold a scalar value with an empirically estimated
    uncertainty from resampling.

    Supports comparison of the values and samples with ``==`` and ``!=``.

    .. rubric:: Examples

    Create a value container with 100 assumed jackknife samples that scatter
    around zero with a standard deviation of 0.1:

    >>> from numpy.random import normal
    >>> samples = normal(loc=0.0, scale=0.01, size=101)
    >>> value = yaw.core.SampledValue(0.0, samples, method="jackknife")
    >>> value
    SampledValue(value=0, error=0.963, n_samples=100, method='jackknife')

    Args:
        value:
            Numerical, scalar value.
        samples (:obj:`NDArray`):
            Samples of ``value`` obtained from resampling methods.
        method (:obj:`str`):
            Resampling method used to obtain the data samples, see
            :class:`~yaw.ResamplingConfig` for available options.
    """

    value: _Tscalar
    """Numerical, scalar value."""
    samples: NDArray[_Tscalar]
    """Samples of ``value`` obtained from resampling methods."""
    method: str
    """Resampling method used to obtain the data samples, see
    :class:`~yaw.ResamplingConfig` for available options."""
    error: _Tscalar = field(init=False)
    """The uncertainty (standard error) of the value."""

    def __post_init__(self) -> None:
        if self.method not in OPTIONS.method:
            raise ValueError(f"unknown sampling method '{self.method}'")
        if self.method == "bootstrap":
            error = np.std(self.samples, ddof=1, axis=0)
        else:  # jackknife
            error = np.std(self.samples, ddof=0, axis=0) * (self.n_samples - 1)
        object.__setattr__(self, "error", error)

    def __repr__(self) -> str:
        string = self.__class__.__name__
        value = self.value
        error = self.error
        n_samples = self.n_samples
        method = self.method
        return f"{string}({value=:.3g}, {error=:.3g}, {n_samples=}, {method=})"

    def __str__(self) -> str:
        return f"{self.value:+.3g}+/-{self.error:.3g}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            if self.samples.shape != other.samples.shape:
                return False
            return (
                self.method == other.method
                and self.value == other.value
                and np.all(self.samples == other.samples)
            )
        else:
            return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    @property
    def n_samples(self) -> int:
        """Number of samples used for error estimate."""
        return len(self.samples)


_Tdata = TypeVar("_Tdata", bound="SampledData")


@dataclass(frozen=True, repr=False)
class SampledData(BinnedQuantity):
    """Container for data and resampled data with redshift binning.

    Contains the redshift binning, data vector, and resampled data vector (e.g.
    jackknife or bootstrap samples). The resampled values are used to compute
    error estimates and covariance/correlation matrices.

    Args:
        binning (:obj:`pandas.IntervalIndex`):
            The redshift binning applied to the data.
        data (:obj:`NDArray`):
            The data values, one for each redshift bin.
        samples (:obj:`NDArray`):
            The resampled data values (e.g. jackknife or bootstrap samples).
        method (:obj:`str`):
            The resampling method used, see :class:`~yaw.ResamplingConfig` for
            available options.

    The container supports addition and subtraction, which return a new instance
    of the container, holding the modified data. This requires that both
    operands are compatible (same binning and same sampling). The operands are
    applied to the ``data`` and ``samples`` attribtes.

    Furthermore, the container supports indexing and iteration over the redshift
    bins using the :obj:`SampledData.bins` attribute. This attribute yields
    instances of :obj:`SampledData` containing a single bin when iterating.
    Slicing and indexing follows the same rules as the underlying ``data``
    :obj:`NDArray`. Refer to :obj:`~yaw.correlation.CorrData` for some indexing
    and iteration examples.

    .. rubric:: Examples

    Create a redshift binning:

    >>> import pandas as pd
    >>> bins = pd.IntervalIndex.from_breaks([0.1, 0.2, 0.3])
    >>> bins
    IntervalIndex([(0.1, 0.2], (0.2, 0.3]], dtype='interval[float64, right]')

    Create some sample data for the bins with value 1 and five assumed jackknife
    samples normal-distributed around 1.

    >>> import numpy as np
    >>> n_bins, n_samples = len(bins), 5
    >>> data = np.ones(n_bins)
    >>> samples = np.random.normal(1.0, size=(n_samples, n_bins))

    Create the container:

    >>> values = yaw.core.SampledData(bins, data, samples, method="jackknife")
    >>> values
    SampledData(n_bins=2, z='0.100...0.300', n_samples=10, method='jackknife')

    Add the container to itself and verify that the values are doubled:

    >>> summed = values + values
    >>> summed.data
    array([2., 2.])

    The same applies to the samples:

    >>> summed.samples / values.samples
    array([[2., 2.],
           [2., 2.],
           [2., 2.],
           [2., 2.],
           [2., 2.]])
    """

    binning: IntervalIndex
    """The redshift bin intervals."""
    data: NDArray
    """The data values, one for each redshift bin."""
    samples: NDArray
    """Samples of the data values, shape (# samples, # bins)."""
    method: str
    """The resampling method used."""
    covariance: NDArray = field(init=False)
    """Covariance matrix automatically computed from the resampled values."""

    def __post_init__(self) -> None:
        if self.data.shape != (self.n_bins,):
            raise ValueError("unexpected shape of 'data' array")
        if not self.samples.shape[1] == self.n_bins:
            raise ValueError("number of bins for 'data' and 'samples' do not match")
        if self.method not in OPTIONS.method:
            raise ValueError(f"unknown sampling method '{self.method}'")

        covmat = cov_from_samples(self.samples, self.method)
        object.__setattr__(self, "covariance", np.atleast_2d(covmat))

    def __repr__(self) -> str:
        string = super().__repr__()[:-1]
        n_samples = self.n_samples
        method = self.method
        return f"{string}, {n_samples=}, {method=})"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            if self.samples.shape != other.samples.shape:
                return False
            return (
                self.method == other.method
                and np.all(self.data == other.data)
                and np.all(self.samples == other.samples)
                and (self.binning == other.binning).all()
            )
        else:
            return False

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __add__(self, other: _Tdata) -> _Tdata:
        self.is_compatible(other, require=True)
        return self.__class__(
            binning=self.get_binning(),
            data=self.data + other.data,
            samples=self.samples + other.samples,
            method=self.method,
        )

    def __sub__(self, other: _Tdata) -> _Tdata:
        self.is_compatible(other, require=True)
        return self.__class__(
            binning=self.get_binning(),
            data=self.data - other.data,
            samples=self.samples - other.samples,
            method=self.method,
        )

    @property
    def bins(self: _Tdata) -> Indexer[int | slice | Sequence, _Tdata]:
        def builder(inst: _Tdata, item: int | slice | Sequence) -> _Tdata:
            if isinstance(item, int):
                item = [item]
            # try to take subsets along bin axis
            binning = inst.get_binning()[item]
            data = inst.data[item]
            samples = inst.samples[:, item]
            # determine which extra attributes need to be copied
            init_attrs = {field.name for field in fields(inst) if field.init}
            copy_attrs = init_attrs - {"binning", "data", "samples"}

            kwargs = dict(binning=binning, data=data, samples=samples)
            kwargs.update({attr: getattr(inst, attr) for attr in copy_attrs})
            return inst.__class__(**kwargs)

        return Indexer(self, builder)

    @property
    def n_samples(self) -> int:
        """Number of samples used for error estimate."""
        return len(self.samples)

    @property
    def error(self) -> NDArray:
        """The uncertainty (standard error) of the data.

        Returns:
            :obj:`NDArray`
        """
        return np.sqrt(np.diag(self.covariance))

    def get_binning(self) -> IntervalIndex:
        return self.binning

    def is_compatible(self, other: SampledData, require: bool = False) -> bool:
        """Check whether this instance is compatible with another instance.

        Ensures that both objects are instances of the same class, that the
        redshift binning is identical, that the number of samples agree, and
        that the resampling method is identical.

        Args:
            other (:obj:`BinnedQuantity`):
                Object instance to compare to.
            require (:obj:`bool`, optional)
                Raise a ValueError if any of the checks fail.

        Returns:
            :obj:`bool`
        """
        if not super().is_compatible(other, require):
            return False
        if self.n_samples != other.n_samples:
            if require:
                raise ValueError("number of samples do not agree")
            return False
        if self.method != other.method:
            if require:
                raise ValueError("resampling method does not agree")
            return False
        return True

    def concatenate_bins(self: _Tdata, *data: _Tdata) -> _Tdata:
        for other in data:
            self.is_compatible(other, require=True)
        all_data: list[_Tdata] = [self, *data]
        binning = concatenate_bin_edges(*all_data)
        # concatenate data
        data = np.concatenate([d.data for d in all_data])
        samples = np.concatenate([d.samples for d in all_data], axis=1)
        # determine which extra attributes need to be copied
        init_attrs = {field.name for field in fields(self) if field.init}
        copy_attrs = init_attrs - {"binning", "data", "samples"}

        kwargs = dict(binning=binning, data=data, samples=samples)
        kwargs.update({attr: getattr(self, attr) for attr in copy_attrs})
        return self.__class__(**kwargs)

    def get_data(self) -> Series:
        """Get the data as :obj:`pandas.Series` with the binning as index."""
        return pd.Series(self.data, index=self.binning)

    def get_samples(self) -> DataFrame:
        """Get the data as :obj:`pandas.DataFrame` with the binning as index.
        The columns are labelled numerically and each represent one of the
        samples."""
        return pd.DataFrame(self.samples.T, index=self.binning)

    def get_error(self) -> Series:
        """Get value error estimate (diagonal of covariance matrix) as series
        with its corresponding redshift bin intervals as index.

        Returns:
            :obj:`pandas.Series`
        """
        return pd.Series(self.error, index=self.binning)

    def get_covariance(self) -> DataFrame:
        """Get value covariance matrix as data frame with its corresponding
        redshift bin intervals as index and column labels.

        Returns:
            :obj:`pandas.DataFrame`
        """
        return pd.DataFrame(
            data=self.covariance, index=self.binning, columns=self.binning
        )

    def get_correlation(self) -> DataFrame:
        """Get value correlation matrix as data frame with its corresponding
        redshift bin intervals as index and column labels.

        Returns:
            :obj:`pandas.DataFrame`
        """
        stdev = np.sqrt(np.diag(self.covariance))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            corr = self.covariance / np.outer(stdev, stdev)
        corr[self.covariance == 0] = 0
        return pd.DataFrame(data=corr, index=self.binning, columns=self.binning)



from __future__ import annotations

from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

import numpy as np
from deprecated import deprecated

from yaw.config import default as DEFAULT
from yaw.config.utils import ConfigError
from yaw.core.abc import DictRepresentation
from yaw.core.cosmology import Scale
from yaw.core.docs import Parameter
from yaw.core.math import array_equal

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray

__all__ = ["ScalesConfig"]


@dataclass(frozen=True)
class ScalesConfig(DictRepresentation):
    """Configuration of scales used for correlation measurements.

    Correlation functions are measured on one or many intervals
    :math:`r_{\\rm min} \\leq r < r_{\\rm max}` angular diameter distance in
    kpc. When measuring correlations, this scale is coverted to angles at the
    current redshift.

    Additionally, pairs can be weighted by their separation
    :math:`r^\\alpha` if a power-law exponent is provided through ``rweight``.
    The weighting is applied logarithmically spaced bins of separation (based
    on the logarithmic bin centers). This is an approximation to actually
    weighting each pair individually and the resolution of this approximation
    can be controlled by setting the number of bins.

    Args:
        rmin (:obj:`float`, :obj:`Sequence[float]`):
            Single or multiple lower scale limits in kpc (angular diameter
            distance).
        rmax (:obj:`float`, :obj:`Sequence[float]`):
            Single or multiple upper scale limits in kpc (angular diameter
            distance).
        rweight (:obj:`float`, optional):
            Power-law exponent used to weight pairs by their separation.
        rbin_num (:obj:`int`, optional):
            Number of radial logarithmic bin used to approximate the weighting
            by separation.
    """

    rmin: Sequence[float] | float = field(
        metadata=Parameter(
            type=float,
            nargs="*",
            required=True,
            help="(list of) lower scale limit in kpc (pyhsical)",
        )
    )
    """Lower scale limit(s) in kpc (angular diameter distance)."""
    rmax: Sequence[float] | float = field(
        metadata=Parameter(
            type=float,
            nargs="*",
            required=True,
            help="(list of) upper scale limit in kpc (pyhsical)",
        )
    )
    """Upper scale limit(s) in kpc (angular diameter distance)."""
    rweight: float | None = field(
        default=DEFAULT.Scales.rweight,
        metadata=Parameter(
            type=float,
            help="weight galaxy pairs by their separation to power 'rweight'",
            default_text="(default: no weighting applied)",
        ),
    )
    """Power-law exponent used to weight pairs by their separation."""
    rbin_num: int = field(
        default=DEFAULT.Scales.rbin_num,
        metadata=Parameter(
            type=int,
            help="number of bins in log r used (i.e. resolution) to compute distance weights",
            default_text="(default: %(default)s)",
        ),
    )
    """Number of radial logarithmic bin used to approximate the weighting by
    separation."""

    def __post_init__(self) -> None:
        msg_scale_error = "scales violates 'rmin' < 'rmax'"
        # validation, set to basic python types
        scalars = (float, int, np.number)
        if isinstance(self.rmin, (Sequence, np.ndarray)) and isinstance(
            self.rmax, (Sequence, np.ndarray)
        ):
            if len(self.rmin) != len(self.rmax):
                raise ConfigError(
                    "number of elements in 'rmin' and 'rmax' do not match"
                )
            # for clean YAML conversion
            for rmin, rmax in zip(self.rmin, self.rmax):
                if rmin >= rmax:
                    raise ConfigError(msg_scale_error)
            if len(self.rmin) == 1:
                rmin = float(self.rmin[0])
                rmax = float(self.rmax[0])
            else:
                rmin = list(float(f) for f in self.rmin)
                rmax = list(float(f) for f in self.rmax)
            object.__setattr__(self, "rmin", rmin)
            object.__setattr__(self, "rmax", rmax)
        elif isinstance(self.rmin, scalars) and isinstance(self.rmax, scalars):
            # for clean YAML conversion
            object.__setattr__(self, "rmin", float(self.rmin))
            object.__setattr__(self, "rmax", float(self.rmax))
            if self.rmin >= self.rmax:
                raise ConfigError(msg_scale_error)
        else:
            raise ConfigError("'rmin' and 'rmax' must be both sequences or float")

    def __eq__(self, other: ScalesConfig) -> bool:
        if not array_equal(self.as_array(), other.as_array()):
            return False
        if self.rweight != other.rweight:
            return False
        if self.rbin_num != other.rbin_num:
            return False
        return True

    def __getitem__(self, idx: int) -> Scale:
        scales = self.as_array()
        return Scale(rmin=scales[idx, 0], rmax=scales[idx, 1])

    def __iter__(self) -> Iterator[Scale]:
        for rmin, rmax in self.as_array():
            yield Scale(rmin=rmin, rmax=rmax)

    def as_array(self) -> NDArray[np.float_]:
        """Obtain the scales cuts as array of shape (2, N)"""
        return np.atleast_2d(np.transpose([self.rmin, self.rmax]))

    @deprecated("use [str(scale) for scale in ScalesConfig] instead", version="2.3.1")
    def dict_keys(self) -> list[str]:
        """Get the scale cuts formatted as a list of strings.

        Format is ``kpc[rmin]t[rmax]``, used as keys to pack outputs of
        correlation measurements in a dictionary when measuring with multiple
        scales cuts.

        .. deprecated:: 2.3.1
            Use instead

            >>> [str(scale) for scale in ScalesConfig]
            ...
        """
        return [str(scale) for scale in self]



from __future__ import annotations

import itertools
import os
import sys
from collections.abc import Iterator
from typing import TYPE_CHECKING, Dict, NoReturn, Tuple

try:  # pragma: no cover
    from typing import TypeAlias
except ImportError:  # pragma: no cover
    from typing_extensions import TypeAlias

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from treecorr import Catalog, NNCorrelation

from yaw.catalogs import BaseCatalog
from yaw.config import Configuration, ResamplingConfig
from yaw.core.coordinates import Coord3D, Coordinate, CoordSky, DistSky
from yaw.core.logging import TimedLog
from yaw.correlation.paircounts import (
    NormalisedCounts,
    PatchedCount,
    PatchedTotal,
    pack_results,
)
from yaw.redshifts import HistData

if TYPE_CHECKING:  # pragma: no cover
    from pandas import DataFrame, Interval

    from yaw.catalogs import PatchLinkage

__all__ = ["EmptyCatalog", "TreecorrCatalog"]


TypeNNResult: TypeAlias = Dict[Tuple[int, int], NNCorrelation]  # supports py3.8


def _iter_bin_masks(
    data: NDArray, bins: NDArray, closed: str = "left"
) -> Iterator[tuple[Interval, NDArray[np.bool_]]]:
    """Split data into bins and return an iterator that yields the boolean masks
    that select the data of the current bin out of the input data array."""
    if closed not in ("left", "right"):
        raise ValueError("'closed' must be either of 'left', 'right'")
    intervals = pd.IntervalIndex.from_breaks(bins, closed=closed)
    bin_ids = np.digitize(data, bins, right=(closed == "right"))
    for i, interval in enumerate(intervals, 1):
        yield interval, (bin_ids == i)


def take_subset(
    cat: TreecorrCatalog, items: NDArray[np.bool_] | NDArray[np.int_] | slice
) -> TreecorrCatalog | EmptyCatalog:
    """Construct a new TreecorrCatalog with a subset of its entries."""
    ra = cat.ra[items]
    if len(ra) == 0:
        return EmptyCatalog(n_patches=cat.n_patches)
    kwargs = dict(
        ra=ra,
        ra_units="radian",
        dec=cat.dec[items],
        dec_units="radian",
        patch=cat.patch[items],
    )
    if cat.has_redshifts():
        kwargs["r"] = cat.redshifts[items]
    if cat.has_weights():
        kwargs["w"] = cat.weights[items]
    return TreecorrCatalog.from_treecorr(Catalog(**kwargs))


class EmptyCatalog:
    """A minimal representation of a TreecorrCatalog that contains no data."""

    def __init__(self, n_patches: int) -> None:
        self.n_patches = n_patches

    def __len__(self) -> int:
        return 0

    def total(self) -> float:
        return 0.0

    def get_totals(self) -> NDArray[np.float_]:
        return np.zeros(self.n_patches)


class TreecorrCatalog(BaseCatalog):
    """An implementation of the :obj:`BaseCatalog` using ``TreeCorr`` for the
    pair counting.

    .. Note::

        The current implementation is not very efficient, because the internal
        fields of the underlying :obj:`treecorr.Catalog` must be rebuilt every
        time the catalog is iterated in redshift bins for the pair counting.

    .. Warning::

        Currently this backend does not support restoration from cache and
        raises an :obj:`NotImplementedError`` when calling the
        :meth:`from_cache` method.
    """

    def __init__(
        self,
        data: DataFrame,
        ra_name: str,
        dec_name: str,
        *,
        patch_name: str | None = None,
        patch_centers: BaseCatalog | Coordinate | None = None,
        n_patches: int | None = None,
        redshift_name: str | None = None,
        weight_name: str | None = None,
        cache_directory: str | None = None,
        progress: bool = False,
    ) -> None:
        # construct the underlying TreeCorr catalogue
        kwargs = dict()
        if cache_directory is not None:
            kwargs["save_patch_dir"] = cache_directory
            if not os.path.exists(cache_directory):
                raise FileNotFoundError(
                    f"patch directory does not exist: '{cache_directory}'"
                )
            self._logger.info(f"using cache directory '{cache_directory}'")

        if n_patches is not None:
            kwargs["npatch"] = n_patches
            log_msg = f"creating {n_patches} patches"
        elif patch_name is not None:
            kwargs["patch"] = data[patch_name]
            log_msg = "splitting data into predefined patches"
        elif isinstance(patch_centers, BaseCatalog):
            kwargs["patch_centers"] = patch_centers.centers.to_3d().values
            n_patches = patch_centers.n_patches
            log_msg = f"applying {n_patches} patches from external data"
        elif isinstance(patch_centers, Coordinate):
            centers = patch_centers.to_3d()
            kwargs["patch_centers"] = centers.values
            n_patches = len(centers)
            log_msg = f"applying {n_patches} patches from external data"
        else:
            raise ValueError(
                "either of 'patch_name', 'patch_centers', or 'n_patches' "
                "must be provided"
            )

        with TimedLog(self._logger.info, log_msg):
            self._catalog = Catalog(
                ra=data[ra_name],
                ra_units="degrees",
                dec=data[dec_name],
                dec_units="degrees",
                r=None if redshift_name is None else data[redshift_name],
                w=None if weight_name is None else data[weight_name],
                **kwargs,
            )
            self._make_patches()

        if cache_directory is not None:
            self.unload()

    @classmethod
    def from_cache(cls, cache_directory: str, progress: bool = False) -> NoReturn:
        """
        Raises:
            NotImplementedError

        .. Warning::

            Currently this backend does not support restoration from cache.
        """
        # super().from_cache(cache_directory)
        # self._make_patches()
        raise NotImplementedError("restoring from cache is currently not supported")

    @classmethod
    def from_treecorr(cls, cat: Catalog) -> TreecorrCatalog:
        """Create a new instace from a :obj:`treecorr.Catalog`."""
        new = cls.__new__(cls)
        new._catalog = cat
        new._make_patches()
        return new

    def _make_patches(self) -> None:
        c = self._catalog
        if c._patches is None:
            low_mem = (not self.is_loaded()) and (c.save_patch_dir is not None)
            c.get_patches(low_mem=low_mem)

    def to_treecorr(self) -> Catalog:
        """Get the internal :obj:`treecorr.Catalog` instance."""
        return self._catalog

    def __len__(self) -> int:
        return self._catalog.ntot

    def __getitem__(self, item: int) -> Catalog:
        return self._catalog._patches[item]

    @property
    def ids(self) -> list[int]:
        return list(range(self.n_patches))

    @property
    def n_patches(self) -> int:
        return self._catalog.npatch

    def __iter__(self) -> Iterator[Catalog]:
        for patch in self._catalog._patches:
            yield patch

    def is_loaded(self) -> bool:
        return self._catalog.loaded

    def load(self) -> None:
        super().load()
        self._catalog.load()

    def unload(self) -> None:
        super().unload()
        self._catalog.unload()

    def has_redshifts(self) -> bool:
        return self.redshifts is not None

    def has_weights(self) -> bool:
        return self.weights is not None

    @property
    def ra(self) -> NDArray[np.float_]:
        return self._catalog.ra

    @property
    def dec(self) -> NDArray[np.float_]:
        return self._catalog.dec

    @property
    def redshifts(self) -> NDArray[np.float_] | None:
        return self._catalog.r

    @property
    def weights(self) -> NDArray[np.float_]:
        return self._catalog.w

    @property
    def patch(self) -> NDArray[np.int_]:
        return self._catalog.patch

    def get_min_redshift(self) -> float:
        if not hasattr(self, "_zmin"):
            if self.has_redshifts():
                self._zmin = self.redshifts.min()
            else:
                self._zmin = None
        return self._zmin

    def get_max_redshift(self) -> float:
        if not hasattr(self, "_zmax"):
            if self.has_redshifts():
                self._zmax = self.redshifts.max()
            else:
                self._zmax = None
        return self._zmax

    @property
    def total(self) -> float:
        return self._catalog.sumw

    def get_totals(self) -> NDArray[np.float_]:
        return np.array([patch.sumw for patch in iter(self)])

    @property
    def centers(self) -> CoordSky:
        centers = Coord3D.from_array(self._catalog.get_patch_centers())
        return centers.to_sky()

    @property
    def radii(self) -> DistSky:
        radii = []
        cls = self.__class__
        for cat in iter(self):
            # build a new TreecorrCatalog without any postprocessing
            patch = cls.__new__(cls)
            patch._catalog = cat
            # compute the angular radius from the maximum separation in 3D
            position = patch.pos.to_3d()
            radius = patch.centers.to_3d().distance(position).max()
            radii.append(radius.to_sky())
        return DistSky.from_dists(radii)

    def iter_bins(
        self, z_bins: NDArray[np.float_], allow_no_redshift: bool = False
    ) -> Iterator[tuple[Interval, TreecorrCatalog | EmptyCatalog]]:
        """Iterate the catalogue in bins of redshift.

        Args:
            z_bins (:obj:`NDArray`):
                Edges of the redshift bins.
            allow_no_redshift (:obj:`bool`):
                If true and the data has no redshifts, the iterator yields the
                whole catalogue at each iteration step.

        Yields:
            (tuple): tuple containing:
                - **intv** (:obj:`pandas.Interval`): the selection for this bin.
                - **cat** (:obj:`TreecorrCatalog`): instance containing the data
                  for this bin.
        """
        if not allow_no_redshift and not self.has_redshifts():
            raise ValueError("no redshifts for iteration provdided")
        if allow_no_redshift:
            for intv in pd.IntervalIndex.from_breaks(z_bins, closed="left"):
                yield intv, self
        else:
            for interval, bin_mask in _iter_bin_masks(self.redshifts, z_bins):
                yield interval, take_subset(self, bin_mask)

    def correlate(
        self,
        config: Configuration,
        binned: bool,
        other: TreecorrCatalog = None,
        linkage: PatchLinkage | None = None,
        progress: bool = False,
    ) -> NormalisedCounts | dict[str, NormalisedCounts]:
        super().correlate(config, binned, other, linkage)

        auto = other is None
        if not auto and not isinstance(other, TreecorrCatalog):
            raise TypeError
        nncorr_config = dict(
            sep_units="radian",
            metric="Arc",
            nbins=(1 if config.scales.rweight is None else config.scales.rbin_num),
            bin_slop=config.backend.rbin_slop,
            num_threads=config.backend.get_threads(),
        )

        # bin the catalogues if necessary
        cats1 = self.iter_bins(config.binning.zbins)
        if auto:
            cats2 = itertools.repeat((None, None))
        else:
            if binned:
                cats2 = other.iter_bins(config.binning.zbins)
            else:
                cats2 = itertools.repeat((None, other))

        # allocate output data containers
        binning = pd.IntervalIndex.from_breaks(config.binning.zbins)
        n_bins = len(binning)
        n_patches = self.n_patches
        totals1 = np.zeros((n_patches, n_bins))
        totals2 = np.zeros((n_patches, n_bins))
        count_dict = {
            str(scale): PatchedCount.zeros(binning, n_patches, auto=auto)
            for scale in config.scales
        }

        # iterate the bins and compute the correlation
        self._logger.debug(
            f"running treecorr on {config.backend.get_threads()} threads"
        )
        for i, ((intv, bincat1), (_, bincat2)) in enumerate(zip(cats1, cats2)):
            if progress:
                _prog_msg = f"processing bin {i+1} / {n_bins}\r"
                sys.stderr.write(_prog_msg)
                sys.stderr.flush()

            angles = [
                scale.to_radian(intv.mid, config.cosmology) for scale in config.scales
            ]
            # extract the total number of objects per patch
            totals1[:, i] = bincat1.get_totals()
            if bincat2 is None:
                totals2[:, i] = totals1[:, i]
            else:
                totals2[:, i] = bincat2.get_totals()

            # trivial case: no data in redshift interval, no counts to update
            if isinstance(bincat1, EmptyCatalog) or isinstance(bincat2, EmptyCatalog):
                continue

            for scale, (ang_min, ang_max) in zip(config.scales, angles):
                # run the correlation measurement
                correlation = NNCorrelation(
                    min_sep=ang_min, max_sep=ang_max, **nncorr_config
                )
                correlation.process(
                    bincat1.to_treecorr(),
                    None if bincat2 is None else bincat2.to_treecorr(),
                )

                # extract the pair counts
                scale_counts = count_dict[str(scale)]
                result: TypeNNResult = correlation.results
                for (pid1, pid2), corr_result in result.items():
                    scale_counts.counts[pid1, pid2, i] = corr_result.weight

        if progress:
            sys.stderr.write((" " * len(_prog_msg)) + "\r")  # clear line

        total = PatchedTotal(  # not scale-dependent
            binning=binning, totals1=totals1, totals2=totals2, auto=auto
        )
        return pack_results(count_dict, total)

    def true_redshifts(
        self,
        config: Configuration,
        sampling_config: ResamplingConfig | None = None,
        progress: bool = False,
    ) -> HistData:
        if sampling_config is None:
            sampling_config = ResamplingConfig()  # default values

        super().true_redshifts(config)
        if not self.has_redshifts():
            raise ValueError("catalog has no redshifts")
        # compute the reshift histogram in each patch
        hist_counts = []
        for patch in iter(self):
            counts, bins = np.histogram(patch.r, config.binning.zbins, weights=patch.w)
            hist_counts.append(counts)
        hist_counts = np.array(hist_counts)

        # construct the output data samples
        binning = pd.IntervalIndex.from_breaks(config.binning.zbins)
        patch_idx = sampling_config.get_samples(self.n_patches)
        nz_data = hist_counts.sum(axis=0)
        nz_samp = np.sum(hist_counts[patch_idx], axis=1)
        return HistData(
            binning=binning,
            data=nz_data,
            samples=nz_samp,
            method=sampling_config.method,
        )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import numpy as np

from yaw.config import OPTIONS
from yaw.config import default as DEFAULT
from yaw.config.utils import ConfigError
from yaw.core.abc import DictRepresentation

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import NDArray

__all__ = ["ResamplingConfig"]


@dataclass(frozen=True)
class ResamplingConfig(DictRepresentation):
    """Configuration for error estimation from spatial resampling.

    Used for all functions and methods that use spatial patches for error
    estimation. Use the :meth:`get_samples` method to generate samples from the
    spatial patches, which can be reused to ensure consistent error estimates
    for different data products that use the same patches.

    Args:
        method (:obj:`str`):
            Resampling method to use, see
            :obj:`~yaw.config.options.Options.method`.
        crosspath (:obj:`str`):
            Whether to use cross-patch pair count measurements.
        n_boot (:obj:`int`):
            Number of samples to generate for the ``bootstrap`` method.
        global_norm (:obj:`bool`):
            Whether to normalise paircounts globally or for each sample. Usually
            not recommended.
        seed (:obj:`int`):
            Random seed to use.
    """

    method: str = DEFAULT.Resampling.method
    """Resampling method to use, see :obj:`~yaw.config.options.Options.method`.
    """
    crosspatch: bool = DEFAULT.Resampling.crosspatch
    """Whether to use cross-patch pair count measurements."""
    n_boot: int = DEFAULT.Resampling.n_boot
    """Number of samples to generate for the ``bootstrap`` method."""
    global_norm: bool = DEFAULT.Resampling.global_norm
    """Whether to normalise paircounts globally or for each sample."""
    seed: int = DEFAULT.Resampling.seed
    """Random seed to use."""
    _resampling_idx: NDArray[np.int_] | None = field(
        default=None, init=False, repr=False
    )

    def __post_init__(self) -> None:
        if self.method not in OPTIONS.method:
            opts = ", ".join(f"'{s}'" for s in OPTIONS.method)
            raise ConfigError(
                f"invalid resampling method '{self.method}', must either of {opts}"
            )

    @property
    def n_patches(self) -> int | None:
        """The number of spatial patches for which this configuratin is valid.

        Available only after generating samples with :meth:`get_samples`.

        Returns:
            int if samples have been generated, else None.
        """
        if self._resampling_idx is None:
            return None
        else:
            return self._resampling_idx.shape[1]

    def _generate_bootstrap(self, n_patches: int) -> NDArray[np.int_]:
        """Generate samples for the bootstrap resampling method.

        For N patches, draw M realisations each containing N randomly chosen
        patches.
        """
        N = n_patches
        rng = np.random.default_rng(seed=self.seed)
        return rng.integers(0, N, size=(self.n_boot, N))

    def _generate_jackknife(self, n_patches: int) -> NDArray[np.int_]:
        """Generate samples for the jackknife resampling method.

        For N patches, draw N realisations by leaving out one of the N patches.
        """
        N = n_patches
        idx = np.delete(np.tile(np.arange(0, N), N), np.s_[:: N + 1])
        return idx.reshape((N, N - 1))

    def get_samples(self, n_patches: int) -> NDArray[np.int_]:
        """Generate a list of patch indices that produces samples for the
        selected resampling method.

        Args:
            n_patches (:obj:`int`):
                Total number of patches for which the samples are generated.

        .. Note::

            Samples are generated only once for each instance. Later calls to
            this method will only check if the number of patches agree with the
            first call and return the initially generated index list. Raises a
            :exc:`ValueError` otherwise.

            The reason is, that the ``bootstrap`` method produces random
            samples, which must be consistent if the resampling is applied to
            different pair count measurements.
        """
        if self._resampling_idx is None:
            if self.method == "jackknife":
                idx = self._generate_jackknife(n_patches)
            else:
                idx = self._generate_bootstrap(n_patches)
            object.__setattr__(self, "_resampling_idx", idx)
        elif n_patches != self.n_patches:
            raise ValueError(
                f"'n_patches' does not match, expected {self.n_patches}, but "
                f"got {n_patches}"
            )
        return self._resampling_idx

    def reset(self) -> None:
        """Reset the internally stored patch indices generated by
        :meth:`get_samples`."""
        object.__setattr__(self, "_resampling_idx", None)

    def to_dict(self) -> dict[str, Any]:
        if self.method == "jackknife":
            return dict(method=self.method, crosspatch=self.crosspatch)
        else:
            return super().to_dict()



"""This module defines simple containers for coordiantes and distances. There
are currently two flavours, 3-dim Euclidean coordinates and distances, as well
as angular coordinates and distances in radian.
"""

from __future__ import annotations

from abc import ABC, abstractclassmethod, abstractmethod, abstractproperty
from collections.abc import Iterator, Sequence
from functools import total_ordering
from typing import TYPE_CHECKING, TypeVar

import numpy as np

from yaw.core.math import sgn

if TYPE_CHECKING:  # pragma: no cover
    from numpy.typing import ArrayLike, NDArray

__all__ = ["Coord3D", "CoordSky", "Dist3D", "DistSky"]


class Coordinate(ABC):
    """Base class for a vector of coordinates."""

    @abstractmethod
    def __init__(self, coords: dict[str, ArrayLike]) -> None:
        pass

    @abstractclassmethod
    def from_array(cls, array) -> Coordinate:
        pass

    @abstractclassmethod
    def from_coords(cls, coords: Sequence[Coordinate]) -> Coordinate:
        """Concatenate a sequence of coordinates into a new vector of
        coordates."""
        pass

    def __repr__(self) -> str:
        pass

    @abstractproperty
    def dim(self) -> tuple[str]:
        """A list of names of coordinates in the coordinate system."""
        pass

    @property
    def ndim(self) -> int:
        """The number of coordinates in the coordinate system."""
        return len(self.dim)

    @abstractproperty
    def values(self) -> NDArray[np.float_]:
        """The coordinate values cast into a numpy array with shape
        `(N, ndim)` or `(ndim,)` if there is only a single entry."""
        pass

    @abstractmethod
    def mean(self) -> Coordinate:
        """The mean coordinate (mean over all dimensions)."""
        pass

    @abstractmethod
    def to_3d(self) -> Coord3D:
        """Get the coordinates as 3-dim Euclidean coordiante :obj:`Coord3D`."""
        pass

    @abstractmethod
    def to_sky(self) -> CoordSky:
        """Get the coordinates as angular coordiante :obj:`CoordSky` in radian.

        .. Warning::
            During conversion, points loose their radial information. After
            back-transformation to :obj:`Coord3D` they lie on the unit sphere.
        """
        pass

    @abstractmethod
    def distance(self, other: Coordinate) -> Distance:
        pass


class Coord3D(Coordinate):
    """A representation of a vector of 3-dim Euclidean coordiantes (x, y, z)."""

    x: NDArray
    """The `x`-coordiante(s)."""
    y: NDArray
    """The `y`-coordiante(s)."""
    z: NDArray
    """The `z`-coordiante(s)."""

    def __init__(self, x: ArrayLike, y: ArrayLike, z: ArrayLike) -> None:
        """Create a new coordinate vector.

        Args:
            x (:obj:`float`, :obj:`NDArray`):
                The `x`-coordiante(s).
            y (:obj:`float`, :obj:`NDArray`):
                The `y`-coordiante(s).
            z (:obj:`float`, :obj:`NDArray`):
                The `z`-coordiante(s).

        .. Note::
            The coordinate vector has a length, supports numpy-style indexing
            and iteration over the elements (always returning new coordinate
            vector instances).
        """
        self.x = np.atleast_1d(x).astype(np.float_)
        self.y = np.atleast_1d(y).astype(np.float_)
        self.z = np.atleast_1d(z).astype(np.float_)

    @classmethod
    def from_array(cls, array) -> Coord3D:
        """Create a new coordinate vector from an array of tuples of
        (`x`, `y`, `z`)."""
        x, y, z = np.transpose(array)
        return cls(x, y, z)

    @classmethod
    def from_coords(cls, coords: Sequence[Coord3D]) -> Coord3D:
        x = np.concatenate([coord.x for coord in coords], axis=0)
        y = np.concatenate([coord.y for coord in coords], axis=0)
        z = np.concatenate([coord.z for coord in coords], axis=0)
        return cls(x, y, z)

    def __repr__(self) -> str:
        x, y, z = self.x, self.y, self.z
        return f"{self.__class__.__name__}({x=}, {y=}, {z=})"

    def __len__(self) -> int:
        return len(self.x)

    def __getitem__(self, idx: ArrayLike) -> Coord3D:
        return self.__class__(x=self.x[idx], y=self.y[idx], z=self.z[idx])

    def __iter__(self) -> Iterator[Coord3D]:
        for i in range(len(self)):
            yield self[i]

    @property
    def dim(self) -> tuple[str]:
        return ("x", "y", "z")

    @property
    def values(self) -> NDArray[np.float_]:
        return np.squeeze(np.transpose([self.x, self.y, self.z]))

    def mean(self) -> Coord3D:
        return Coord3D(x=self.x.mean(), y=self.y.mean(), z=self.z.mean())

    def to_3d(self) -> Coord3D:
        return self

    def to_sky(self) -> CoordSky:
        x = self.x
        y = self.y
        z = self.z
        r_d2 = np.sqrt(x * x + y * y)
        r_d3 = np.sqrt(x * x + y * y + z * z)
        # transform
        ra = np.arccos(x / r_d2) * sgn(y) % (2.0 * np.pi)
        ra[np.isnan(ra)] = 0.0
        dec = np.arcsin(self.z / r_d3)
        return CoordSky(ra, dec)

    def distance(self, other: Coordinate) -> Dist3D:
        """Compute the Euclidean distance between two coordinate vectors.

        Coordinates are automatically converted before distance calculation.

        Args:
            other (:obj:`Coordinate`):
                Second coordinate vector.

        Returns:
            :obj:`Dist3D`:
                Euclidean distance between points in this and the other vector.
        """
        c1 = self.to_3d()
        c2 = other.to_3d()
        return Dist3D(
            np.sqrt((c1.x - c2.x) ** 2 + (c1.y - c2.y) ** 2 + (c1.z - c2.z) ** 2)
        )


class CoordSky(Coordinate):
    """A representation of a vector of angular coordinates in radian.

    Angles follow the astronomical convention of R.A./Dec., i.e. declination
    values are in the range of [:math:`-\\pi`, :math:`\\pi`].
    """

    ra: NDArray
    """The right ascension coordinate(s)."""
    dec: NDArray
    """The declination coordinate(s)."""

    def __init__(self, ra: ArrayLike, dec: ArrayLike) -> None:
        """Create a new coordinate vector.

        Args:
            ra (:obj:`float`, :obj:`NDArray`):
                The right ascension coordinate(s).
            dec (:obj:`float`, :obj:`NDArray`):
                The declination coordinate(s).

        .. Note::
            The coordinate vector has a length, supports numpy-style indexing
            and iteration over the elements (always returning new coordinate
            vector instances).
        """
        self.ra = np.atleast_1d(ra).astype(np.float_)
        self.dec = np.atleast_1d(dec).astype(np.float_)

    @classmethod
    def from_array(cls, array):
        """Create a new coordinate vector from an array of tuples of
        (`ra`, `dec`)."""
        ra, dec = np.transpose(array)
        return cls(ra, dec)

    @classmethod
    def from_coords(cls, coords: Sequence[CoordSky]) -> CoordSky:
        ra = np.concatenate([coord.ra for coord in coords], axis=0)
        dec = np.concatenate([coord.dec for coord in coords], axis=0)
        return cls(ra, dec)

    def __repr__(self) -> str:
        ra, dec = self.ra, self.dec
        return f"{self.__class__.__name__}({ra=}, {dec=})"

    def __len__(self) -> int:
        return len(self.ra)

    def __getitem__(self, idx: ArrayLike) -> CoordSky:
        return self.__class__(ra=self.ra[idx], dec=self.dec[idx])

    def __iter__(self) -> Iterator[CoordSky]:
        for i in range(len(self)):
            yield self[i]

    @property
    def dim(self) -> tuple[str]:
        return ("ra", "dec")

    @property
    def values(self) -> NDArray[np.float_]:
        return np.squeeze(np.transpose([self.ra, self.dec]))

    def mean(self) -> Coord3D:
        return self.to_3d().mean().to_sky()

    def to_3d(self) -> Coord3D:
        cos_dec = np.cos(self.dec)
        return Coord3D(
            x=np.cos(self.ra) * cos_dec, y=np.sin(self.ra) * cos_dec, z=np.sin(self.dec)
        )

    def to_sky(self) -> CoordSky:
        return self

    def distance(self, other: Coordinate) -> DistSky:
        """Compute the angular distance in radian between two coordinate
        vectors.

        Coordinates are automatically converted before distance calculation.

        Args:
            other (:obj:`Coordinate`):
                Second coordinate vector.

        Returns:
            :obj:`DistSky`:
                Angular distance in radian between points in this and the other
                vector.
        """
        # lazy shortcut
        self_3D = self.to_3d()
        other_3D = other.to_3d()
        dist = self_3D.distance(other_3D)
        return dist.to_sky()


_Tdist = TypeVar("_Tdist", bound="Distance")


@total_ordering
class Distance(ABC):
    """Base class for a vector of distances."""

    def __init__(self, distance: ArrayLike) -> None:
        """Constructs a new vector of distances.

        Args:
            distance (:obj:`float`, :obj:`NDArray`):
                Distance values.

        .. Note::
            The coordinate vector has a length, supports numpy-style indexing
            and iteration over the elements (always returning new coordinate
            vector instances).

            Additionally, distances implement element-wise addition and
            subtraction, as well as the comparison operators.
        """
        self._values = np.atleast_1d(distance).astype(np.float_)

    @classmethod
    def from_dists(cls: _Tdist, dists: Sequence[_Tdist]) -> _Tdist:
        """Concatenate a sequence of distances into a new vector of distances."""
        return cls(np.concatenate([dist._values for dist in dists], axis=0))

    def __len__(self) -> int:
        return len(self._values)

    def __getitem__(self, idx: ArrayLike) -> CoordSky:
        return self.__class__(self._values[idx])

    def __iter__(self) -> Iterator[CoordSky]:
        for i in range(len(self)):
            yield self[i]

    @property
    def values(self) -> NDArray[np.float_] | float:
        """The distances cast into a numpy array with shape `(N,)` or a scalar
        if there is only a single distance."""
        return np.squeeze(self._values)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.values})"

    def __format__(self, __format_spec: str) -> str:
        return self.values.__format__(__format_spec)

    def __eq__(self, other: Distance) -> ArrayLike[np.bool_]:
        return self.values == other.values

    def __lt__(self, other: Distance) -> ArrayLike[np.bool_]:
        return self.values < other.values

    @abstractmethod
    def __add__(self: _Tdist, other: _Tdist) -> _Tdist:
        pass

    @abstractmethod
    def __sub__(self: _Tdist, other: _Tdist) -> _Tdist:
        pass

    def min(self) -> _Tdist:
        """Compute the minimum value and return it as new `Distance`
        instance."""
        return self.__class__(self.values.min())

    def max(self) -> _Tdist:
        """Compute the maximum value and return it as new `Distance`
        instance."""
        return self.__class__(self.values.max())

    @abstractmethod
    def to_3d(self) -> Dist3D:
        """Convert the distance to the Euclidean distance."""
        pass

    @abstractmethod
    def to_sky(self) -> DistSky:
        """Convert the distance to an angular separation in radian."""
        pass


class Dist3D(Distance):
    """Implements a vector of Euclidean distances."""

    def __add__(self, other: Dist3D) -> DistSky:
        dist_sky = self.to_sky() + other.to_sky()
        return dist_sky.to_3d()

    def __sub__(self, other: Dist3D) -> DistSky:
        dist_sky = self.to_sky() - other.to_sky()
        return dist_sky.to_3d()

    def to_3d(self) -> Dist3D:
        return self

    def to_sky(self) -> DistSky:
        if np.any(self._values > 2.0):
            raise ValueError("distance exceeds size of unit sphere")
        return DistSky(2.0 * np.arcsin(self.values / 2.0))


class DistSky(Distance):
    """Implements a vector of angular distances in radian."""

    def __add__(self, other: DistSky) -> DistSky:
        return DistSky(self.values + other.values)

    def __sub__(self, other: DistSky) -> DistSky:
        return DistSky(self.values - other.values)

    def to_3d(self) -> Dist3D:
        return Dist3D(2.0 * np.sin(self.values / 2.0))

    def to_sky(self) -> DistSky:
        return self



from __future__ import annotations

import os
from dataclasses import dataclass, field

from yaw.config import default as DEFAULT
from yaw.core.abc import DictRepresentation
from yaw.core.docs import Parameter

__all__ = ["BackendConfig"]


@dataclass(frozen=True)
class BackendConfig(DictRepresentation):
    """Configuration of backends used for correlation measurements.

    Args:
        thread_num (:obj:`int`, optional):
            Number of threads to use for parallel processing.
        crosspatch (:obj:`bool`, optional):
            Whether to count pairs across patch boundaries (``scipy`` backend
            only).
        rbin_slop (:obj:`int`, optional):
            `TreeCorr` ``rbin_slop`` parameter (``treecorr`` backend only).
    """

    # general
    thread_num: int | None = field(
        default=DEFAULT.Backend.thread_num,
        metadata=Parameter(
            type=int,
            help="default number of threads to use",
            default_text="(default: all)",
        ),
    )
    """Number of threads to use for parallel processing."""
    # scipy
    crosspatch: bool = field(
        default=DEFAULT.Backend.crosspatch,
        metadata=Parameter(
            type=bool,
            help="whether to count pairs across patch boundaries (scipy backend only)",
        ),
    )
    """Whether to count pairs across patch boundaries (``scipy`` backend only).
    """
    # treecorr
    rbin_slop: float = field(
        default=DEFAULT.Backend.rbin_slop,
        metadata=Parameter(
            type=float,
            help="TreeCorr 'rbin_slop' parameter",
            default_text="(default: %(default)s), without 'rweight' this just "
            "a single radial bin, otherwise 'rbin_num'",
        ),
    )
    """`TreeCorr` ``rbin_slop`` parameter (``treecorr`` backend only)."""

    def __post_init__(self) -> None:
        if self.thread_num is None:
            object.__setattr__(self, "thread_num", os.cpu_count())

    def get_threads(self, max=None) -> int:
        """Get the number of threads for parallel processing.

        The value is capped at the number of logical cores available.
        """
        if self.thread_num is None:
            thread_num = os.cpu_count()
        else:
            thread_num = self.thread_num
        if max is not None:
            if max < 1:
                raise ValueError("'max' must be positive")
            thread_num = min(max, thread_num)
        return thread_num



"""This module implements some abstract base classes that define the interfaces
for high level containers in other modules.
"""

from __future__ import annotations

from abc import ABC, abstractclassmethod, abstractmethod, abstractproperty
from dataclasses import asdict
from typing import TYPE_CHECKING, Any, Type, TypeVar

import h5py
import numpy as np
import pandas as pd
from numpy.typing import NDArray

if TYPE_CHECKING:  # pragma: no cover
    from pandas import IntervalIndex

    from yaw.core.containers import Indexer
    from yaw.core.utils import TypePathStr

__all__ = [
    "PatchedQuantity",
    "BinnedQuantity",
    "concatenate_bin_edges",
    "HDFSerializable",
    "DictRepresentation",
]


_Tpatched = TypeVar("_Tpatched", bound="PatchedQuantity")


class PatchedQuantity(ABC):
    """Base class for an object that has data organised in spatial patches."""

    @abstractproperty
    def n_patches(self) -> int:
        """Get the number of spatial patches."""
        pass

    @abstractproperty
    def patches(self) -> Indexer:
        """An :obj:`~yaw.core.containers.Indexer` attribute that supports
        iteration over the spatial patches or selecting a subset of the patches.

        The indexer always returns new container instances with the indexed
        data subset or the current item when iterating.

        .. Note::
            Indexing rules for a one-dimensional numpy array apply.

        Returns:
            :obj:`yaw.core.containers.Indexer`
        """
        raise NotImplementedError

    @abstractmethod
    def concatenate_patches(self: _Tpatched, *data: _Tpatched) -> _Tpatched:
        """Concatenate pair count data containers with equal redshift binning.

        The data is merged by extending the dimension of the patch axes. The
        resulting data array will be a block matrix of the input data arrays,
        i.e. all elements with correlations between different inputs set to
        zero.

        .. Note::
            Necessary condition for merging is that the the redshift binning of
            all inputs is identical. Cannot merge cross- with autocorrelation
            containers.

        Args:
            *data:
                Containers of same type that are appended to the patch dimension
                of this container.

        Returns:
            New instance of this container with combined data.
        """
        raise NotImplementedError


_Tbinned = TypeVar("_Tbinned", bound="BinnedQuantity")


class BinnedQuantity(ABC):
    """Base class for an object that has data organised in redshift bins."""

    def get_binning(self) -> IntervalIndex:
        """Get the underlying, exact redshift bin intervals.

        Returns:
            :obj:`pandas.IntervalIndex`
        """
        raise NotImplementedError

    def __repr__(self) -> str:
        name = self.__class__.__name__
        n_bins = self.n_bins
        binning = self.get_binning()
        z = f"{binning[0].left:.3f}...{binning[-1].right:.3f}"
        return f"{name}({n_bins=}, {z=})"

    @property
    def n_bins(self) -> int:
        """Get the number of redshift bins."""
        return len(self.get_binning())

    @property
    def mids(self) -> NDArray[np.float_]:
        """Get the centers of the redshift bins as array."""
        return np.array([z.mid for z in self.get_binning()])

    @property
    def edges(self) -> NDArray[np.float_]:
        """Get the edges of the redshift bins as flat array."""
        binning = self.get_binning()
        return np.append(binning.left, binning.right[-1])

    @property
    def dz(self) -> NDArray[np.float_]:
        """Get the width of the redshift bins as array."""
        return np.diff(self.edges)

    @property
    def closed(self) -> str:
        """Specifies on which side the redshift bin intervals are closed, can
        be: ``left``, ``right``, ``both``, ``neither``."""
        return self.get_binning().closed

    @abstractproperty
    def bins(self) -> Indexer:
        """An :obj:`~yaw.core.containers.Indexer` attribute that supports
        iteration over the bins or selecting a subset of the bins.

        The indexer always returns new container instances with the indexed
        data subset or the current item when iterating.

        .. Warning::
            Indexing rules for a one-dimensional numpy array apply, however if
            the resulting binning is not contiguous or contains repeated bins,
            some operations on the returned container may fail.

        Returns:
            :obj:`yaw.core.containers.Indexer`
        """
        raise NotImplementedError

    def is_compatible(self: _Tbinned, other: _Tbinned, require: bool = False) -> bool:
        """Check whether this instance is compatible with another instance.

        Ensures that both objects are instances of the same class and that the
        redshift binning is identical.

        Args:
            other (:obj:`BinnedQuantity`):
                Object instance to compare to.
            require (:obj:`bool`, optional)
                Raise a ValueError if any of the checks fail.

        Returns:
            :obj:`bool`
        """
        if not isinstance(other, self.__class__):
            raise TypeError(
                f"object of type {type(other)} is not compatible with "
                f"{self.__class__}"
            )
        if self.n_bins != other.n_bins:
            if require:
                raise ValueError("number of bins do not agree")
            return False
        if np.any(self.get_binning() != other.get_binning()):
            if require:
                raise ValueError("binning is not identical")
            return False
        return True

    @abstractmethod
    def concatenate_bins(self: _Tbinned, *data: _Tbinned) -> _Tbinned:
        """Concatenate pair count data containers with equal patches.

        The data is merged by appending the data along the redshift binning
        axis.

        .. Note::
            Necessary condition for merging is that the patch numbers are
            identical and that the merged binning is contiguous and
            non-overlapping. Cannot merge cross- with autocorrelation
            containers.

        Args:
            *data:
                Containers of same type that are appended to the patch dimension
                of this container.

        Returns:
            New instance of this container with combined data.
        """
        raise NotImplementedError


def concatenate_bin_edges(*patched: BinnedQuantity) -> IntervalIndex:
    """Concatenate the binning a set of data containers.

    The input containers are automatically sorted by the lowest edge of the
    redshift binning. Necessary condidtions for mergning are are that the patch
    numbers are identical and that the resulting is contiguous and
    non-overlapping, i.e. the final edge of the previous binning must be
    identical to the lowest edge of the next binning.
    """
    patched = sorted([p for p in patched], key=lambda p: p.edges[0])
    reference = patched[0]
    edges = reference.edges
    for other in patched[1:]:
        if edges[-1] == other.edges[0]:
            edges = np.concatenate([edges, other.edges[1:]])
        else:
            raise ValueError("cannot merge, bins are not contiguous")
    return pd.IntervalIndex.from_breaks(edges, closed=reference.closed)


_Thdf = TypeVar("_Thdf", bound="HDFSerializable")


class HDFSerializable(ABC):
    """Base class for an object that can be serialised into a HDF5 file."""

    @abstractclassmethod
    def from_hdf(cls: Type[_Thdf], source: h5py.Group) -> _Thdf:
        """Create a class instance by deserialising data from a HDF5 group.

        Args:
            source (:obj:`h5py.Group`):
                Group in an opened HDF5 file that contains the serialised data.

        Returns:
            :obj:`HDFSerializablep`
        """
        raise NotImplementedError

    @abstractmethod
    def to_hdf(self, dest: h5py.Group) -> None:
        """Serialise the class instance into an existing HDF5 group.

        Args:
            dest (:obj:`h5py.Group`):
                Group in which the serialised data structures are created.
        """
        raise NotImplementedError

    @classmethod
    def from_file(cls: Type[_Thdf], path: TypePathStr) -> _Thdf:
        """Create a class instance by deserialising data from a HDF5 file.

        Args:
            path (:obj:`pathlib.Path`, :obj:`str`):
                Group in an opened HDF5 file that contains the necessary data.

        Returns:
            :obj:`HDFSerializable`
        """
        with h5py.File(str(path)) as f:
            return cls.from_hdf(f)

    def to_file(self, path: TypePathStr) -> None:
        """Serialise the class instance to a new HDF5 file.

        Args:
            path (:obj:`pathlib.Path`, :obj:`str`):
                Path at which the HDF5 file is created.
        """
        with h5py.File(str(path), mode="w") as f:
            self.to_hdf(f)


_Tdict = TypeVar("_Tdict", bound="DictRepresentation")


class DictRepresentation(ABC):
    """Base class for an object that can be serialised into a dictionary."""

    @classmethod
    def from_dict(
        cls: Type[_Tdict],
        the_dict: dict[str, Any],
        **kwargs: dict[str, Any],  # passing additional constructor data
    ) -> _Tdict:
        """Create a class instance from a dictionary representation of the
        minimally required data.

        Args:
            the_dict (:obj:`dict`):
                Dictionary containing the data.
            **kwargs: Additional data needed to construct the class instance.
        """
        return cls(**the_dict)

    def to_dict(self) -> dict[str, Any]:
        """Serialise the class instance to a dictionary containing a minimal set
        of required data.

        Returns:
            :obj:`dict`
        """
        return asdict(self)
